import numpy as np
import pytest
from pandas import DataFrame, Index, MultiIndex, Series
from xlwings import Sheet

from xlviews.core.range import Range
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def df():
    df = DataFrame(
        {
            "x": [1, 1, 1, 1, 2, 2, 2, 2],
            "y": [1, 1, 2, 2, 1, 1, 2, 2],
            "a": [1, 2, 3, 4, 5, 6, 7, 8],
            "b": [11, 12, 13, 14, 15, 16, 17, 18],
        },
    )
    return df.set_index(["x", "y"])


@pytest.fixture
def sf(df: DataFrame, sheet: Sheet):
    return SheetFrame(2, 2, data=df, sheet=sheet)


@pytest.mark.parametrize(
    ("func", "value"),
    [("sum", 36), ("min", 1), ("max", 8), ("mean", 4.5)],
)
def test_df_str(df: DataFrame, func: str, value: float):
    s = df.agg(func)
    assert isinstance(s, Series)
    assert s.index.to_list() == ["a", "b"]
    assert s["a"] == value


def test_df_dict(df: DataFrame):
    s = df.agg({"a": "min", "b": "max"})
    assert isinstance(s, Series)
    assert s.index.to_list() == ["a", "b"]
    np.testing.assert_array_equal(s, [1, 18])


def test_df_dict_one(df: DataFrame):
    s = df.agg({"a": "min"})
    assert isinstance(s, Series)
    assert s.index.to_list() == ["a"]
    np.testing.assert_array_equal(s, [1])


def test_df_list(df: DataFrame):
    x = df.agg(["min", "max"])
    assert isinstance(x, DataFrame)
    assert x.index.to_list() == ["min", "max"]
    assert x.columns.to_list() == ["a", "b"]
    np.testing.assert_array_equal(x, [[1, 11], [8, 18]])


@pytest.mark.parametrize("func", ["sum", "count", "min", "max", "mean"])
def test_sf_str(sf: SheetFrame, df: DataFrame, func: str):
    a = sf.agg(func, formula=True)
    b = df.agg(func)
    assert isinstance(a, Series)
    assert a.index.to_list() == b.index.to_list()
    sf = SheetFrame(20, 2, data=a.to_frame(), sheet=sf.sheet)
    np.testing.assert_array_equal(sf.value[func], b)


@pytest.mark.parametrize("name", ["sum", "count", "min", "max", "mean"])
def test_sf_range(sf: SheetFrame, df: DataFrame, name: str):
    func = Range((20, 1), sheet=sf.sheet)
    a = sf.agg(func, formula=True)
    b = df.agg(name)
    assert isinstance(a, Series)
    assert a.index.to_list() == b.index.to_list()
    sf = SheetFrame(20, 2, data=a.to_frame(), sheet=sf.sheet)
    func.value = name
    np.testing.assert_array_equal(sf.value[0], b)


def test_sf_str_columns(sf: SheetFrame):
    a = sf.agg("mean", columns="a", formula=True)
    assert len(a) == 1
    sf = SheetFrame(20, 2, data=a.to_frame(), sheet=sf.sheet)
    np.testing.assert_array_equal(sf.value, [[4.5]])


def test_sf_str_columns_list(sf: SheetFrame):
    a = sf.agg("mean", columns=["a", "b"], formula=True)
    assert len(a) == 2
    sf = SheetFrame(20, 2, data=a.to_frame(), sheet=sf.sheet)
    df = DataFrame([[4.5], [14.5]], index=["a", "b"], columns=["mean"])
    df_sf = sf.value
    assert df_sf.equals(df)
    assert df_sf.index.equals(df.index)
    assert df_sf.columns.equals(df.columns)


def test_sf_dict(sf: SheetFrame, df: DataFrame):
    func = {"a": "min", "b": "max"}
    a = sf.agg(func, formula=True)
    b = df.agg(func)
    assert isinstance(a, Series)
    assert a.index.to_list() == b.index.to_list()
    sf = SheetFrame(20, 2, data=a.to_frame(), sheet=sf.sheet)
    np.testing.assert_array_equal(sf.value[0], b)


def test_sf_list(sf: SheetFrame, df: DataFrame):
    func = ["min", "max"]
    a = sf.agg(func, formula=True)
    b = df.agg(func)  # type: ignore
    assert isinstance(a, DataFrame)
    assert a.index.to_list() == b.index.to_list()
    assert a.columns.to_list() == b.columns.to_list()
    sf = SheetFrame(20, 2, data=a, sheet=sf.sheet)
    np.testing.assert_array_equal(sf.value, b)


def test_sf_list_columns(sf: SheetFrame, df: DataFrame):
    a = sf.agg(["sum", "count"], columns="b", formula=True)
    assert isinstance(a, DataFrame)
    sf = SheetFrame(20, 2, data=a, sheet=sf.sheet)
    np.testing.assert_array_equal(sf.value, [[116], [8]])


def test_sf_none(sf: SheetFrame):
    s = sf.agg(None)
    assert isinstance(s, Series)
    assert s["a"] == "$D$3:$D$10"
    assert s["b"] == "$E$3:$E$10"


def test_sf_first(sf: SheetFrame):
    s = sf.agg("first", formula=True)
    assert isinstance(s, Series)
    assert s["a"] == "=$D$3"
    assert s["b"] == "=$E$3"


def test_sf_first_none(sf: SheetFrame):
    s = sf.agg({"x": "first", "b": None}, formula=True)
    assert isinstance(s, Series)
    assert len(s) == 2
    assert s["x"] == "=$B$3"
    assert s["b"] == "=$E$3:$E$10"


def test_df_group_str_str(df: DataFrame):
    a = df.groupby("x").agg("sum")
    assert a.columns.to_list() == ["a", "b"]
    assert a.index.to_list() == [1, 2]
    assert a.index.name == "x"
    np.testing.assert_array_equal(a, [[10, 50], [26, 66]])


@pytest.mark.filterwarnings("ignore::FutureWarning")
def test_df_group_str_str_as_index_false(df: DataFrame):
    a = df.groupby("x", as_index=False).agg("sum")
    assert a.columns.to_list() == ["a", "b"]
    assert a.index.to_list() == [0, 1]
    assert a.index.name is None
    np.testing.assert_array_equal(a, [[10, 50], [26, 66]])


def test_df_group_list_str(df: DataFrame):
    a = df.groupby(["x", "y"]).agg("sum")
    assert a.columns.to_list() == ["a", "b"]
    assert a.index.to_list() == [(1, 1), (1, 2), (2, 1), (2, 2)]
    assert a.index.names == ["x", "y"]
    np.testing.assert_array_equal(a, [[3, 23], [7, 27], [11, 31], [15, 35]])


def test_df_group_list_list(df: DataFrame):
    a = df.groupby(["x", "y"]).agg(["min", "max"])
    c = [("a", "min"), ("a", "max"), ("b", "min"), ("b", "max")]
    assert a.columns.to_list() == c
    assert a.index.to_list() == [(1, 1), (1, 2), (2, 1), (2, 2)]
    assert a.index.names == ["x", "y"]
    v = [[1, 2, 11, 12], [3, 4, 13, 14], [5, 6, 15, 16], [7, 8, 17, 18]]
    np.testing.assert_array_equal(a, v)


def test_df_group_list_dict(df: DataFrame):
    a = df.groupby(["x", "y"]).agg({"a": "min", "b": "max"})
    assert a.columns.to_list() == ["a", "b"]
    assert a.index.to_list() == [(1, 1), (1, 2), (2, 1), (2, 2)]
    assert a.index.names == ["x", "y"]
    np.testing.assert_array_equal(a, [[1, 12], [3, 14], [5, 16], [7, 18]])


def test_index_str(sf: SheetFrame):
    a = sf.groupby("x").index()
    b = Index([1, 2], name="x")
    assert a.equals(b)


def test_index_list(sf: SheetFrame):
    a = sf.groupby(["x", "y"]).index()
    b = MultiIndex.from_tuples([(1, 1), (1, 2), (2, 1), (2, 2)], names=["x", "y"])
    assert a.equals(b)


def test_index_str_as_address(sf: SheetFrame):
    a = sf.groupby("x").index(as_address=True)
    b = Index(["$B$3", "$B$7"], name="x")
    assert a.equals(b)


def test_index_list_as_address(sf: SheetFrame):
    a = sf.groupby(["x", "y"]).index(as_address=True, formula=True)
    values = [(f"=$B${r}", f"=$C${r}") for r in [3, 5, 7, 9]]
    b = MultiIndex.from_tuples(values, names=["x", "y"])
    assert a.equals(b)


@pytest.mark.parametrize("func", ["sum", "median", "mean"])
@pytest.mark.parametrize("by", ["x", "y"])
def test_sf_group_str_str(sf: SheetFrame, df: DataFrame, func, by):
    a = sf.groupby(by).agg(func, as_address=True, formula=True)
    b = df.groupby(by).agg(func).astype(float)
    sf = SheetFrame(50, 2, data=a, sheet=sf.sheet)
    df_sf = sf.value
    assert df_sf.equals(b)
    assert df_sf.index.equals(b.index)
    assert df_sf.columns.equals(b.columns)


@pytest.mark.parametrize("func", ["sum", "median", "mean"])
@pytest.mark.parametrize("by", ["x", "y"])
def test_sf_group_str_range(sf: SheetFrame, df: DataFrame, func, by):
    rng = Range((50, 1), sheet=sf.sheet)
    rng.value = func
    a = sf.groupby(by).agg(rng, as_address=True, formula=True)
    b = df.groupby(by).agg(func).astype(float)
    sf = SheetFrame(50, 2, data=a, sheet=sf.sheet)
    df_sf = sf.value
    assert df_sf.equals(b)
    assert df_sf.index.equals(b.index)
    assert df_sf.columns.equals(b.columns)


@pytest.mark.parametrize("func", ["sum", "min", "max"])
@pytest.mark.parametrize("by", [["x", "y"], ["y", "x"]])
def test_sf_group_list_str(sf: SheetFrame, df: DataFrame, func, by):
    a = sf.groupby(by).agg(func, as_address=True, formula=True)
    b = df.groupby(by).agg(func).astype(float)
    sf = SheetFrame(50, 10, data=a, sheet=sf.sheet)
    df_sf = sf.value
    assert df_sf.equals(b)
    assert df_sf.index.equals(b.index)
    assert df_sf.columns.equals(b.columns)


@pytest.mark.parametrize("by", [["x", "y"], ["y", "x"]])
@pytest.mark.parametrize("sort", [True, False])
def test_sf_group_list_str_sort(sf: SheetFrame, df: DataFrame, by, sort):
    a = sf.groupby(by, sort=sort).agg("sum", as_address=True, formula=True)
    b = df.groupby(by, sort=sort).agg("sum").astype(float)
    sf = SheetFrame(50, 20, data=a, sheet=sf.sheet)
    df_sf = sf.value
    assert df_sf.equals(b)
    assert df_sf.index.equals(b.index)
    assert df_sf.columns.equals(b.columns)


@pytest.mark.parametrize(
    "func",
    [{"a": "min", "b": "max"}, {"a": "median", "b": "mean"}],
)
@pytest.mark.parametrize("by", [["x", "y"], ["y", "x"]])
@pytest.mark.parametrize("sort", [True, False])
def test_sf_group_list_dict(sf: SheetFrame, df: DataFrame, func, by, sort):
    a = sf.groupby(by, sort=sort).agg(func, as_address=True, formula=True)
    b = df.groupby(by, sort=sort).agg(func).astype(float)
    sf = SheetFrame(50, 30, data=a, sheet=sf.sheet)
    df_sf = sf.value
    assert df_sf.equals(b)
    assert df_sf.index.equals(b.index)
    assert df_sf.columns.equals(b.columns)


@pytest.mark.parametrize("func", [["sum", "mean"], ["min", "max"]])
@pytest.mark.parametrize("sort", [True, False])
def test_sf_group_list_list(sf: SheetFrame, df: DataFrame, func, sort):
    by = ["y", "x"]
    a = sf.groupby(by, sort=sort).agg(func, as_address=True, formula=True)
    b = df.groupby(by, sort=sort).agg(func).astype(float)
    sf = SheetFrame(50, 40, data=a, sheet=sf.sheet)
    df_sf = sf.value
    assert df_sf.equals(b)
    assert df_sf.index.equals(b.index)
    assert df_sf.columns.equals(b.columns)



from __future__ import annotations

from typing import TYPE_CHECKING

from xlviews.core.range import Range
from xlviews.dataframes.heat_frame import Colorbar, HeatFrame
from xlviews.testing.common import create_sheet
from xlviews.testing.sheet_frame.pivot import Pivot

if TYPE_CHECKING:
    from collections.abc import Hashable, Iterator
    from typing import Any

    from xlviews.dataframes.sheet_frame import SheetFrame


def facet(sf: SheetFrame) -> Iterator[tuple[dict[Hashable, Any], HeatFrame]]:
    sf.set_adjacent_column_width(1)

    rng = sf.get_range("u")
    cb = Colorbar(3, 11, 12)
    cb.set(vmin=rng, vmax=rng).autofit()
    cb.set_adjacent_column_width(1)

    df = sf.pivot_table("u", ["B", "Y"], ["A", "X"], aggfunc="mean", formula=True)

    for key, frame in HeatFrame.facet(2, 13, df, index="B"):
        frame.autofit()
        frame.set_adjacent_column_width(1)
        cb.apply(frame.range)
        cell = Range(frame.row - 1, frame.column + 1)
        cell.value = str(key)
        yield key, frame


if __name__ == "__main__":
    sheet = create_sheet()
    fc = Pivot(sheet, style=True)
    list(facet(fc.sf))



from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING

import xlwings
from xlwings.constants import AxisType, ChartType, Placement, TickMark

from xlviews.config import rcParams
from xlviews.style import set_font_api
from xlviews.utils import suspend_screen_updates

from .series import Series
from .style import (
    get_axis_label,
    get_axis_scale,
    get_ticks,
    set_area_format,
    set_axis_label,
    set_axis_scale,
    set_dimensions,
    set_tick_labels,
    set_ticks,
)

if TYPE_CHECKING:
    from typing import Any, Self

    from xlwings import Chart, Sheet


def chart_position(
    sheet: Sheet,
    left: float | None,
    top: float | None,
) -> tuple[float, float]:
    """Return the position of the chart."""
    if left and left > 0 and top and top > 0:
        return left, top

    if not sheet.charts:
        return left or rcParams["chart.left"], top or rcParams["chart.top"]

    if left == 0:
        for chart in sheet.charts:
            left = chart.left if left == 0 else min(left, chart.left)
        if top and top > 0:
            return left, top
        top = 0
        for chart in sheet.charts:
            top = max(top, chart.top + chart.height)
        return left, top

    if top == 0:
        for chart in sheet.charts:
            top = chart.top if top == 0 else min(top, chart.top)
        if left and left > 0:
            return left, top
        left = 0
        for chart in sheet.charts:
            left = max(left, chart.left + chart.width)
        return left, top

    chart = sheet.charts[-1]
    if left is None and top is None:
        return chart.left + chart.width, chart.top

    return chart.left, chart.top + chart.height


class Axes:
    sheet: Sheet
    chart: Chart
    chart_type: int
    series_collection: list[Series]

    @suspend_screen_updates
    def __init__(
        self,
        row: int | None = None,
        column: int | None = None,
        chart_type: int = ChartType.xlXYScatter,
        sheet: Sheet | None = None,
        *,
        left: float | None = None,
        top: float | None = None,
        width: float = 0,
        height: float = 0,
        border_width: int = 0,
        visible_only: bool = True,
        has_legend: bool = True,
        include_in_layout: bool = False,
    ) -> None:
        self.sheet = sheet or xlwings.sheets.active

        if row:
            top = self.sheet.range(row, 1).top
        if column:
            left = self.sheet.range(1, column).left

        left, top = chart_position(self.sheet, left, top)

        width = width or rcParams["chart.width"]
        height = height or rcParams["chart.height"]

        self.chart = self.sheet.charts.add(left, top, width, height)  # type: ignore

        self.chart_type = chart_type
        self.chart.api[1].ChartType = chart_type

        # self.chart.api[0].Placement = xw.constants.Placement.xlMove
        self.chart.api[0].Placement = Placement.xlFreeFloating
        self.chart.api[0].Border.LineStyle = border_width
        self.chart.api[1].PlotVisibleOnly = visible_only

        self.xaxis.MajorTickMark = TickMark.xlTickMarkInside
        self.yaxis.MajorTickMark = TickMark.xlTickMarkInside

        self.chart.api[1].HasLegend = has_legend
        self.chart.api[1].Legend.IncludeInLayout = include_in_layout

        self.series_collection = []

    @suspend_screen_updates
    def copy(
        self,
        row: int | None = None,
        column: int | None = None,
        *,
        left: float | None = None,
        top: float | None = None,
    ) -> Self:
        border_width = self.chart.api[0].Border.LineStyle
        visible_only = self.chart.api[1].PlotVisibleOnly
        has_legend = self.chart.api[1].HasLegend
        include_in_layout = self.chart.api[1].Legend.IncludeInLayout

        if left == 0:
            left = self.chart.left + self.chart.width
            top = self.chart.top

        if top == 0:
            left = self.chart.left
            top = self.chart.top + self.chart.height

        return self.__class__(
            row=row,
            column=column,
            chart_type=self.chart_type,
            sheet=self.sheet,
            left=left,
            top=top,
            width=self.chart.width,
            height=self.chart.height,
            border_width=border_width,
            visible_only=visible_only,
            has_legend=has_legend,
            include_in_layout=include_in_layout,
        )

    @property
    def xaxis(self):  # noqa: ANN201
        chart = self.chart.api[1]
        return chart.Axes(AxisType.xlCategory)

    @property
    def yaxis(self):  # noqa: ANN201
        chart = self.chart.api[1]
        return chart.Axes(AxisType.xlValue)

    @suspend_screen_updates
    def add_series(
        self,
        x: Any,
        y: Any | None = None,
        label: str | None = None,
        chart_type: int | None = None,
    ) -> Series:
        if chart_type is None:
            chart_type = self.chart_type

        series = Series(self, x, y, label, chart_type)
        self.series_collection.append(series)

        return series

    @property
    def title(self) -> str | None:
        api = self.chart.api[1]

        if api.HasTitle:
            return api.ChartTitle.Text

        return None

    @title.setter
    def title(self, value: str | None) -> None:
        self.set_title(value)

    def set_title(
        self,
        title: str | None = None,
        *,
        name: str | None = None,
        size: int | None = None,
        **kwargs,
    ) -> None:
        api = self.chart.api[1]

        if title is None:
            api.HasTitle = False
            return

        api.HasTitle = True
        chart_title = api.ChartTitle
        chart_title.Text = title

        name = name or rcParams["chart.font.name"]
        size = size or rcParams["chart.title.font.size"]
        set_font_api(chart_title, name, size=size, **kwargs)

    @property
    def xlabel(self) -> str | None:
        return get_axis_label(self.xaxis)

    @xlabel.setter
    def xlabel(self, value: str | None) -> None:
        self.set_xlabel(value)

    @property
    def ylabel(self) -> str | None:
        return get_axis_label(self.yaxis)

    @ylabel.setter
    def ylabel(self, value: str | None) -> None:
        self.set_ylabel(value)

    def set_xlabel(self, label: str | None = None, **kwargs) -> None:
        set_axis_label(self.xaxis, label, **kwargs)

    def set_ylabel(self, label: str | None = None, **kwargs) -> None:
        set_axis_label(self.yaxis, label, **kwargs)

    @property
    def xticks(self) -> tuple[float, float, float, float]:
        return get_ticks(self.xaxis)

    @xticks.setter
    def xticks(self, value: tuple[float, ...]) -> None:
        set_ticks(self.xaxis, *value)

    @property
    def yticks(self) -> tuple[float, float, float, float]:
        return get_ticks(self.yaxis)

    @yticks.setter
    def yticks(self, value: tuple[float, ...]) -> None:
        set_ticks(self.yaxis, *value)

    def set_xticks(self, *args, **kwargs) -> None:
        set_ticks(self.xaxis, *args, **kwargs)

    def set_yticks(self, *args, **kwargs) -> None:
        set_ticks(self.yaxis, *args, **kwargs)

    def set_xtick_labels(self, *args, **kwargs) -> None:
        set_tick_labels(self.xaxis, *args, **kwargs)

    def set_ytick_labels(self, *args, **kwargs) -> None:
        set_tick_labels(self.yaxis, *args, **kwargs)

    @property
    def xscale(self) -> str:
        return get_axis_scale(self.xaxis)

    @xscale.setter
    def xscale(self, scale: str) -> None:
        set_axis_scale(self.xaxis, scale)

    @property
    def yscale(self) -> str:
        return get_axis_scale(self.yaxis)

    @yscale.setter
    def yscale(self, scale: str) -> None:
        set_axis_scale(self.yaxis, scale)

    @suspend_screen_updates
    def set(
        self,
        xlabel: str | None = "",
        ylabel: str | None = "",
        xticks: tuple[float, ...] | None = None,
        yticks: tuple[float, ...] | None = None,
        xscale: str | None = None,
        yscale: str | None = None,
        title: str | None = None,
        style: bool = True,
        tight_layout: bool = True,
        legend: bool | tuple[float, float] = False,
    ) -> Self:
        if xlabel != "":
            self.xlabel = xlabel
        if ylabel != "":
            self.ylabel = ylabel
        if xticks:
            self.xticks = xticks
        if yticks:
            self.yticks = yticks
        if xscale:
            self.xscale = xscale
        if yscale:
            self.yscale = yscale
        if title:
            self.title = title
        if style:
            self.style()
        if tight_layout:
            self.tight_layout()

        if isinstance(legend, tuple):
            self.legend(loc=legend)
        elif legend:
            self.legend()

        return self

    def delete_legend(self) -> None:
        api = self.chart.api[1]
        if api.HasLegend:
            api.Legend.Delete()

    @suspend_screen_updates
    def legend(
        self,
        left: float | None = None,
        top: float | None = None,
        width: float | None = None,
        height: float | None = None,
        *,
        name: str | None = None,
        size: int | None = None,
        border: str | int | None = None,
        fill: str | int | None = None,
        alpha: float | None = None,
        loc: tuple[float, float] | None = (1, 1),
        margin: float = 3,
        entry_height_scale: float = 1,
    ) -> Self:
        self.delete_legend()
        api = self.chart.api[1]
        api.HasLegend = True

        legend = api.Legend
        legend.IncludeInLayout = False

        legend_entries = list(legend.LegendEntries())
        it = zip(legend_entries, self.series_collection, strict=True)
        for entry, series in it:
            if not series.label:
                entry.Delete()

        if api.HasLegend is False:
            return self

        name = name or rcParams["chart.font.name"]
        size = size or rcParams["chart.legend.font.size"]
        set_font_api(legend, name, size=size)

        if height is None:
            heights = [0]
            for entry in legend.LegendEntries():
                with suppress(Exception):
                    heights.append(entry.Height * entry_height_scale)
            height = sum(heights)

        if width is None:
            widths = [0]
            for entry in legend.LegendEntries():
                with suppress(Exception):
                    widths.append(entry.Width)
            width = max(widths)

        set_dimensions(legend, left, top, width, height)

        if border is None:
            border = rcParams["chart.legend.border.color"]
        if fill is None:
            fill = rcParams["chart.legend.fill.color"]
        if alpha is None:
            alpha = rcParams["chart.legend.fill.alpha"]
        set_area_format(legend, border, fill, alpha)

        if loc:
            x, y = loc
            x = (x + 1) / 2
            y = (1 - y) / 2

            plot_area = api.PlotArea
            inside_left = plot_area.InsideLeft + margin
            inside_top = plot_area.InsideTop + margin
            inside_width = plot_area.InsideWidth - 2 * margin
            inside_height = plot_area.InsideHeight - 2 * margin

            left = inside_left + x * inside_width - x * legend.Width
            top = inside_top + y * inside_height - y * legend.Height

            set_dimensions(legend, left, top)

        return self

    def tight_layout(self, title_height_scale: float = 0.7) -> Self:
        api = self.chart.api[1]
        xaxis = self.xaxis
        yaxis = self.yaxis

        if not (api.HasTitle and xaxis.HasTitle and yaxis.HasTitle):
            return self

        title = api.ChartTitle
        pa = api.PlotArea
        ga = self.chart.api[0]

        title.Top = 0
        yaxis.AxisTitle.Left = 0
        xaxis.AxisTitle.Top = ga.Height - xaxis.AxisTitle.Height

        pa.Top = title_height_scale * title.Height
        pa.Left = yaxis.AxisTitle.Width
        pa.Width = ga.Width - pa.Left
        pa.Height = ga.Height - pa.Top - xaxis.AxisTitle.Height

        title.Left = pa.InsideLeft + pa.InsideWidth / 2 - title.Width / 2

        xaxis.AxisTitle.Left = (
            pa.InsideLeft + pa.InsideWidth / 2 - xaxis.AxisTitle.Width / 2
        )
        yaxis.AxisTitle.Top = (
            pa.InsideTop + pa.InsideHeight / 2 - yaxis.AxisTitle.Height / 2
        )

        return self

    def style(self) -> Self:
        # msoElementPrimaryCategoryGridLinesMajor = 334
        api = self.chart.api[1]
        api.SetElement(334)
        # msoElementPrimaryValueGridLinesMajor == 330
        api.SetElement(330)

        plot_area = api.PlotArea
        line = plot_area.Format.Line
        line.Visible = True
        line.ForeColor.RGB = 0
        line.Weight = 1.2
        line.Transparency = 0.5

        line = self.xaxis.MajorGridlines.Format.Line
        line.Visible = True
        line.ForeColor.RGB = 0
        line.Weight = 1
        line.Transparency = 0.7

        line = self.yaxis.MajorGridlines.Format.Line
        line.Visible = True
        line.ForeColor.RGB = 0
        line.Weight = 1
        line.Transparency = 0.7

        return self



import pytest
from xlwings import Sheet

from xlviews.chart.axes import Axes
from xlviews.figure.grid import Grid, Series
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def seriesx(sheet_module: Sheet):
    ax = Axes(left=30, top=40, width=120, height=150, sheet=sheet_module)
    return Series(ax, 3, axis=0)


def test_series_len(seriesx: Series):
    assert len(seriesx) == 3


@pytest.mark.parametrize(
    ("k", "left", "top"),
    [(0, 30, 40), (1, 150, 40), (2, 270, 40)],
)
def test_seriesx_init(seriesx: Series, k: int, left, top):
    assert seriesx[k].chart.left == left
    assert seriesx[k].chart.top == top
    assert seriesx[:][k].chart.left == left
    assert seriesx[:][k].chart.top == top


@pytest.fixture(scope="module")
def seriesy(sheet_module: Sheet):
    ax = Axes(left=30, top=40, width=120, height=150, sheet=sheet_module)
    return Series(ax, 3, axis=1)


@pytest.mark.parametrize(
    ("k", "left", "top"),
    [(0, 30, 40), (1, 30, 190), (2, 30, 340)],
)
def test_seriesy_init(seriesy: Series, k: int, left, top):
    assert seriesy[k].chart.left == left
    assert seriesy[k].chart.top == top
    assert seriesy[:][k].chart.left == left
    assert seriesy[:][k].chart.top == top


def test_series_getitem_error(seriesx: Series):
    with pytest.raises(ValueError, match="Invalid key: abc"):
        seriesx["abc"]  # type: ignore


def test_series_iter(seriesx: Series):
    ax = list(seriesx)[-1]
    assert ax.chart.left == 270
    assert ax.chart.top == 40


def test_series_axis_error(seriesx):
    ax = seriesx[0]
    with pytest.raises(ValueError, match="Invalid axis: 2"):
        Series(ax, 1, axis=2)  # type: ignore


@pytest.fixture(scope="module")
def grid(sheet_module: Sheet):
    ax = Axes(left=30, top=40, width=120, height=150, sheet=sheet_module)
    return Grid(ax, 3, 4)


@pytest.mark.parametrize(
    ("r", "c", "left", "top"),
    [
        (0, 0, 30, 40),
        (0, 1, 150, 40),
        (0, 2, 270, 40),
        (0, 3, 390, 40),
        (1, 0, 30, 190),
        (1, 1, 150, 190),
        (1, 2, 270, 190),
        (1, 3, 390, 190),
        (2, 0, 30, 340),
        (2, 1, 150, 340),
        (2, 2, 270, 340),
        (2, 3, 390, 340),
    ],
)
def test_grid_init(grid: Grid, r: int, c: int, left, top):
    assert grid[r, c].chart.left == left
    assert grid[r, c].chart.top == top
    assert grid[r][c].chart.left == left
    assert grid[r][c].chart.top == top
    assert grid[:, c][r].chart.left == left
    assert grid[:, c][r].chart.top == top
    assert grid[r, :][c].chart.left == left
    assert grid[r, :][c].chart.top == top
    assert grid[:, :][r, c].chart.left == left


def test_grid_shape(grid: Grid):
    assert grid.shape == (3, 4)


def test_grid_len(grid: Grid):
    assert len(grid) == 3


def test_grid_shape_empty():
    grid = Grid([])
    assert grid.shape == (0, 0)


def test_grid_getitem_error(grid: Grid):
    with pytest.raises(ValueError, match="Invalid key: abc"):
        grid["abc"]  # type: ignore


def test_grid_iter(grid: Grid):
    assert next(iter(grid))[-1].chart.left == 390
    assert next(iter(grid))[-1].chart.top == 40



import pytest
from xlwings import Sheet

from xlviews.dataframes.heat_frame import HeatFrame
from xlviews.testing import is_app_available
from xlviews.testing.heat_frame.agg import AggParent, AggRange, AggStr

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc_parent(sheet_module: Sheet):
    return AggParent(sheet_module)


@pytest.fixture(scope="module")
def sf_str(fc_parent: AggParent):
    return AggStr(fc_parent.sf).sf


def test_index(sf_str: HeatFrame):
    assert sf_str.sheet.range("I3:I6").value == [1, 2, 3, 4]


def test_columns(sf_str: HeatFrame):
    assert sf_str.sheet.range("J2:L2").value == [1, 2, 3]


@pytest.mark.parametrize(
    ("i", "value"),
    [(3, [8.5, 80.5, 152.5]), (4, [26.5, 98.5, 170.5])],
)
def test_values(sf_str: HeatFrame, i: int, value: int):
    assert sf_str.sheet.range(f"J{i}:L{i}").value == value


@pytest.fixture(scope="module")
def sf_range(fc_parent: AggParent):
    return AggRange(fc_parent.sf).sf


@pytest.mark.parametrize(
    ("func", "value"),
    [("min", 0), ("max", 17), ("mean", 8.5), ("count", 18)],
)
def test_values_func(sf_range: HeatFrame, func, value):
    sf_range.sheet.range("$N$13").value = func
    assert sf_range.sheet.range("J9").value == value



import numpy as np
import pytest
from pandas import DataFrame, Series
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import FrameContainer, is_app_available
from xlviews.testing.sheet_frame.base import WideColumn

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return WideColumn(sheet_module, 4, 2)


@pytest.fixture(scope="module")
def sf(fc: FrameContainer):
    sf = fc.sf
    sf.add_formula_column("u", "={u}+{a}")
    sf.add_formula_column("v", "={v}+{b}")
    return sf


@pytest.fixture(scope="module")
def df(sf: SheetFrame):
    return sf.value


@pytest.mark.parametrize("func", ["sum", "max", "mean"])
def test_str(sf: SheetFrame, df: DataFrame, func: str):
    a = sf.agg(func, formula=True)
    b = df.agg(func)
    assert isinstance(a, Series)
    sf = SheetFrame(20, 2, a.to_frame(), sf.sheet)
    np.testing.assert_array_equal(sf.value[func], b)


def test_list(sf: SheetFrame, df: DataFrame):
    func = ["min", "max", "median", "sum"]
    a = sf.agg(func, formula=True)
    b = df.agg(func)  # type: ignore
    assert isinstance(a, DataFrame)
    assert isinstance(b, DataFrame)
    assert a.index.to_list() == b.index.to_list()
    sf = SheetFrame(50, 2, a, sf.sheet)
    np.testing.assert_array_equal(sf.value, b)


def test_sf_none(sf: SheetFrame):
    s = sf.agg(None)
    assert isinstance(s, Series)
    assert s["a"] == "$D$5:$D$9"
    assert s["b"] == "$E$5:$E$9"


def test_sf_first(sf: SheetFrame):
    s = sf.agg("first", formula=True)
    assert isinstance(s, Series)
    assert s["a"] == "=$D$5"
    assert s["b"] == "=$E$5"



"""DataFrame on an Excel sheet."""

from __future__ import annotations

import re
from functools import partial
from typing import TYPE_CHECKING, overload

import numpy as np
import pandas as pd
import xlwings
from pandas import DataFrame, Series
from xlwings import Range as RangeImpl
from xlwings import Sheet
from xlwings.constants import Direction

from xlviews.core.address import index_to_column_name
from xlviews.core.formula import Func, aggregate
from xlviews.core.index import Index
from xlviews.core.range import Range, iter_addresses
from xlviews.style import set_alignment
from xlviews.utils import suspend_screen_updates

from .groupby import GroupBy
from .style import set_frame_style, set_wide_column_style
from .table import Table

if TYPE_CHECKING:
    from collections.abc import Iterator, Sequence
    from typing import Any, Literal, Self

    from numpy.typing import NDArray


class SheetFrame:
    """Data frame on an Excel sheet."""

    cell: RangeImpl
    sheet: Sheet
    index: pd.Index
    columns: Index
    table: Table | None = None

    @suspend_screen_updates
    def __init__(
        self,
        row: int,
        column: int,
        data: DataFrame,
        sheet: Sheet | None = None,
    ) -> None:
        """Create a DataFrame on an Excel sheet.

        Args:
            row (int): The row index of the top-left cell.
            column (int): The column index of the top-left cell.
            data (DataFrame): The DataFrame to write to the sheet.
            sheet (Sheet, optional): The sheet object.
        """
        self.sheet = sheet or xlwings.sheets.active
        self.cell = self.sheet.range(row, column)

        self.index = data.index
        self.columns = Index(data.columns)

        self.cell.options(DataFrame).value = data

        if data.columns.nlevels > 1 and data.index.nlevels == 1:
            self.cell.options(transpose=True).value = data.columns.names

    def expand(self) -> Range:
        start = self.row, self.column
        end = start[0] + self.height - 1, start[1] + self.width - 1
        return Range(start, end, sheet=self.sheet)

    def __repr__(self) -> str:
        rng = self.expand()
        cls = self.__class__.__name__
        return f"<{cls} {rng.get_address(external=True)}>"

    def __len__(self) -> int:
        return len(self.index)

    def __contains__(self, item: Any) -> bool:
        if not isinstance(item, str | tuple):
            return False

        return item in self.columns

    def __iter__(self) -> Iterator[str | tuple[str, ...] | None]:
        return iter(self.columns)

    @property
    def row(self) -> int:
        """Return the row of the top-left cell."""
        return self.cell.row

    @property
    def column(self) -> int:
        """Return the column of the top-left cell."""
        return self.cell.column

    @property
    def shape(self) -> tuple[int, int]:
        return len(self), len(self.columns)

    @property
    def height(self) -> int:
        return self.columns.nlevels + len(self)

    @property
    def width(self) -> int:
        return self.index.nlevels + len(self.columns)

    @property
    def value(self) -> DataFrame:
        rng = self.expand().impl
        index = self.index.nlevels
        header = self.columns.nlevels
        df = rng.options(DataFrame, index=index, header=header).value

        if not isinstance(df, DataFrame):
            raise NotImplementedError

        return df

    def get_loc(self, column: str) -> int | tuple[int, int]:
        if column in self.index.names:
            return self.index.names.index(column) + self.column

        return self.columns.get_loc(column, self.column + self.index.nlevels)

    @overload
    def get_indexer(self, columns: list[str] | None) -> list[int]: ...

    @overload
    def get_indexer(self, columns: dict[str, Any]) -> NDArray[np.intp]: ...

    def get_indexer(
        self,
        columns: list[str] | dict[str, Any] | None = None,
        **kwargs,
    ) -> list[int] | NDArray[np.intp]:
        if isinstance(columns, dict) or (columns is None and kwargs):
            return self.columns.get_indexer(
                columns,
                self.column + self.index.nlevels,
                **kwargs,
            )

        column = self.column
        if columns is None:
            columns = self.columns.to_list()
            start = column + self.index.nlevels
            end = start + len(columns)
            return list(range(start, end))

        cs = [*self.index.names, *self.columns]
        return [cs.index(c) + column for c in columns]

    @overload
    def get_range(
        self,
        columns: str,
        offset: Literal[0, -1] | None = None,
    ) -> Range: ...

    @overload
    def get_range(
        self,
        columns: list[str] | None,
        offset: Literal[0, -1] | None = None,
    ) -> list[Range]: ...

    def get_range(
        self,
        columns: str | list[str] | None,
        offset: Literal[0, -1] | None = None,
    ) -> Range | list[Range]:
        if self.columns.nlevels != 1:
            raise NotImplementedError

        match offset:
            case 0:
                start = end = self.row + 1
            case -1:
                start = end = self.row
            case None:
                start = self.row + 1
                end = start + len(self) - 1
            case _:
                msg = f"invalid offset: {offset}"
                raise ValueError(msg)

        if isinstance(columns, str):
            loc = self.get_loc(columns)
            if isinstance(loc, int):
                loc = loc, loc
            return Range((start, loc[0]), (end, loc[1]), self.sheet)

        idx = self.get_indexer(columns)
        return [Range((start, i), (end, i), sheet=self.sheet) for i in idx]

    def iter_ranges(self, axis: Literal[0, 1] = 0) -> Iterator[Range]:
        if axis == 0:
            start = self.row + self.columns.nlevels
            end = start + len(self) - 1
            offset = self.column + self.index.nlevels

            for index in range(len(self.columns)):
                yield Range((start, index + offset), (end, index + offset), self.sheet)

        elif axis == 1:
            start = self.column + self.index.nlevels
            end = start + len(self.columns) - 1
            offset = self.row + self.columns.nlevels

            for index in range(len(self)):
                yield Range((index + offset, start), (index + offset, end), self.sheet)

        else:
            raise ValueError("axis must be 0 or 1")

    def add_column(
        self,
        column: str,
        value: Any | None = None,
        *,
        number_format: str | None = None,
        autofit: bool = False,
        style: bool = False,
    ) -> None:
        if self.columns.nlevels != 1:
            raise NotImplementedError

        index = self.column + self.width
        self.sheet.range(self.row, index).value = column
        self.columns.append(column)

        end = self.row + len(self)
        rng = self.sheet.range((self.row + 1, index), (end, index))

        if value is not None:
            rng.options(transpose=True).value = value
            if number_format:
                rng.number_format = number_format

        if autofit:
            rng = self.sheet.range((self.row, index), (end, index))
            rng.autofit()

        if style:
            self.style()

    def add_wide_column(
        self,
        column: str,
        values: Any,
        *,
        number_format: str | None = None,
        autofit: bool = False,
        style: bool = False,
    ) -> None:
        """Add a wide column.

        Args:
            column (str): The name of the wide column.
            values (iterable): The values to be expanded horizontally.
            number_format (str, optional): The number format.
            autofit (bool): Whether to autofit the width.
            style (bool): Whether to style the column.
        """
        if self.columns.nlevels != 1:
            raise NotImplementedError

        index = self.column + self.width
        rng = self.sheet.range(self.row - 1, index)
        rng.value = column
        set_alignment(rng, horizontal_alignment="center")
        self.columns.append(column, values)

        rng = self.sheet.range((self.row, index), (self.row, index + len(values)))
        rng.value = values
        if number_format:
            rng.number_format = number_format

        if autofit:
            rng.autofit()

        if style:
            self.style()

    def add_formula_column(
        self,
        column: str,
        formula: str,
        *,
        number_format: str | None = None,
        autofit: bool = False,
        style: bool = False,
    ) -> None:
        """Add a formula column.

        Args:
            rng (Range): The range of the column.
            formula (str or tuple): The formula.
            number_format (str, optional): The number format.
            autofit (bool): Whether to autofit the width.
        """
        if self.columns.nlevels != 1:
            raise NotImplementedError

        refs = {}
        for m in re.finditer(r"{(.+?)}", formula):
            key = m.group(1)
            loc = self.get_loc(key)

            if isinstance(loc, int):
                ref = Range(self.row + 1, loc, self.sheet)
                addr = ref.get_address(row_absolute=False)

            else:
                ref = Range((self.row, loc[0]), (self.row, loc[0]), self.sheet)
                addr = ref.get_address(column_absolute=False)

            refs[key] = addr

        if isinstance(column, str) and column not in self.columns:
            self.add_column(column)

        rng = self.get_range(column).impl
        rng.value = formula.format(**refs)

        if number_format:
            rng.number_format = number_format

        if autofit:
            self.sheet.range(rng[0].offset(-1), rng[-1]).autofit()

        if style:
            self.style()

    @overload
    def get_address(
        self,
        columns: str,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> Series: ...

    @overload
    def get_address(
        self,
        columns: list[str] | None = None,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> DataFrame: ...

    def get_address(
        self,
        columns: str | list[str] | None = None,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> Series | DataFrame:
        if isinstance(columns, str):
            columns = [columns]
            is_str = True
        else:
            is_str = False

        rngs = self.get_range(columns)

        if columns is None:
            columns = self.columns.to_list()

        agg = partial(
            iter_addresses,
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
            cellwise=True,
            formula=formula,
        )

        values = np.array([list(agg(r)) for r in rngs]).T
        df = DataFrame(values, index=self.index, columns=columns)

        return df[columns[0]] if is_str else df

    @overload
    def agg(
        self,
        func: Func | dict = None,
        columns: str | list[str] | None = None,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> Series: ...

    @overload
    def agg(
        self,
        func: Sequence[Func],
        columns: str | list[str] | None = None,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> DataFrame: ...

    def agg(
        self,
        func: Func | dict | Sequence[Func] = None,
        columns: str | list[str] | None = None,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> Series | DataFrame:
        if self.columns.nlevels != 1:
            if isinstance(func, Func) and columns is None:
                return self.melt(
                    func,
                    row_absolute=row_absolute,
                    column_absolute=column_absolute,
                    include_sheetname=include_sheetname,
                    external=external,
                    formula=formula,
                )

            raise NotImplementedError

        if isinstance(func, dict):
            columns = list(func.keys())
        elif isinstance(columns, str):
            columns = [columns]

        if isinstance(func, Range | RangeImpl):
            if func.sheet.book.name != self.sheet.book.name:
                raise ValueError("Range is from a different book")
            if func.sheet.name != self.sheet.name:
                raise ValueError("Range is from a different sheet")

        rngs = self.get_range(columns)

        if columns is None:
            columns = self.columns.to_list()

        agg = partial(
            self._agg,
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
            formula=formula,
        )

        if isinstance(func, dict):
            it = zip(rngs, func.values(), strict=True)
            return Series([agg(f, r) for r, f in it], index=columns)

        if func is None or isinstance(func, str | Range | RangeImpl):
            name = func if isinstance(func, str) else None
            return Series([agg(func, r) for r in rngs], index=columns, name=name)

        values = [[agg(f, r) for r in rngs] for f in func]
        return DataFrame(values, index=list(func), columns=columns)

    def _agg(self, func: Func, rng: Range, **kwargs) -> str:
        if func == "first":
            rng = rng[0]
            func = None

        return aggregate(func, rng, **kwargs)

    def melt(
        self,
        func: Func = None,
        value_name: str = "value",
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> Series | DataFrame:
        """Unpivot a SheetFrame from wide to long format."""
        agg = partial(
            aggregate,
            func,
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
            formula=formula,
        )

        index = self.columns.nlevels == 1
        df = self.columns.to_frame(index=index)
        df[value_name] = list(map(agg, self.iter_ranges(axis=0)))

        if index:
            s = df[value_name]
            if isinstance(func, str | None):
                s.name = func
            return s

        return df

    def pivot_table(
        self,
        values: str | list[str] | None = None,
        index: str | list[str] | None = None,
        columns: str | list[str] | None = None,
        aggfunc: Func | list[Func] = None,
        row_absolute: bool = True,
        column_absolute: bool = True,
        include_sheetname: bool = False,
        external: bool = False,
        formula: bool = False,
    ) -> DataFrame:
        if isinstance(aggfunc, list):
            dfs = [
                self.pivot_table(
                    values,
                    index,
                    columns,
                    f,
                    row_absolute,
                    column_absolute,
                    include_sheetname,
                    external,
                    formula,
                )
                for f in aggfunc
            ]

            keys = [f if isinstance(f, str) else None for f in aggfunc]
            return pd.concat(dfs, axis=1, keys=keys)

        if aggfunc is None:
            data = self.get_address(
                [values] if isinstance(values, str) else values,
                row_absolute=row_absolute,
                column_absolute=column_absolute,
                include_sheetname=include_sheetname,
                external=external,
                formula=formula,
            )

        if index is None:
            by = []
        else:
            by = [index] if isinstance(index, str) else index

        if columns is None:
            if not by:
                raise ValueError("No group keys passed!")
        else:
            by = [*by, columns] if isinstance(columns, str) else by + columns

        data = self.groupby(by).agg(
            aggfunc,
            values,
            row_absolute=row_absolute,
            column_absolute=column_absolute,
            include_sheetname=include_sheetname,
            external=external,
            formula=formula,
        )

        return data.pivot_table(values, index, columns, aggfunc=lambda x: x)

    def groupby(self, by: str | list[str] | None, *, sort: bool = True) -> GroupBy:
        return GroupBy(self, by, sort=sort)

    def get_number_format(self, column: str) -> str:
        idx = self.get_loc(column)
        if isinstance(idx, tuple):
            idx = idx[0]
        return self.sheet.range(self.row + self.columns.nlevels, idx).number_format

    def number_format(
        self,
        number_format: str | dict | None = None,
        *,
        autofit: bool = False,
        **columns_format,
    ) -> Self:
        if isinstance(number_format, dict):
            columns_format.update(number_format)

        row_start = self.row + self.columns.nlevels
        row_end = row_start + len(self) - 1

        if self.columns.nlevels == 1:
            for column in [*self.index.names, *self.columns]:
                if not column:
                    continue

                for pattern, number_format in columns_format.items():
                    if re.match(pattern, column):
                        rng = self.get_range(column).impl
                        rng.number_format = number_format
                        if autofit:
                            rng.autofit()
                        break

        elif isinstance(number_format, str):
            for i in self.get_indexer(columns_format):
                rng = self.sheet.range((row_start, i), (row_end, i))
                rng.number_format = number_format
                if autofit:
                    rng.autofit()

        else:
            raise NotImplementedError

        return self

    def style(self, **kwargs) -> Self:
        set_frame_style(self, **kwargs)
        set_wide_column_style(self)
        return self

    def autofit(self) -> Self:
        start = self.cell
        end = start.offset(self.height - 1, self.width - 1)
        self.sheet.range(start, end).autofit()
        return self

    def alignment(self, alignment: str) -> Self:
        start = self.cell
        end = start.offset(0, self.width - 1)
        rng = self.sheet.range(start, end)
        set_alignment(rng, alignment)
        return self

    def set_adjacent_column_width(self, width: float) -> None:
        """Set the width of the adjacent empty column."""
        column = self.column + self.width
        self.sheet.range(1, column).column_width = width

    def get_adjacent_cell(self, offset: int = 0) -> RangeImpl:
        """Get the adjacent cell of the SheetFrame."""
        return self.cell.offset(0, self.width + offset + 1)

    def move(self, count: int, direction: str = "down", width: int = 0) -> None:
        return move(self, count, direction, width)

    def as_table(
        self,
        *,
        const_header: bool = True,
        autofit: bool = True,
        style: bool = True,
    ) -> Table:
        if self.table:
            return self.table

        if self.columns.nlevels != 1:
            raise NotImplementedError

        self.alignment("left")

        end = self.cell.offset(len(self), self.width - 1)
        rng = self.sheet.range(self.cell, end)

        table = Table(
            rng,
            autofit=autofit,
            const_header=const_header,
            style=style,
            index_nlevels=self.index.nlevels,
        )
        self.table = table

        return table

    def unlist(self) -> Self:
        if self.table:
            self.table.unlist()
            self.table = None

        return self


def move(sf: SheetFrame, count: int, direction: str = "down", width: int = 0) -> None:
    """Insert empty rows/columns to move the SheetFrame to the right or down.

    Args:
        count (int): The number of empty rows/columns to insert.
        direction (str): 'down' or 'right'
        width (int, optional): The width of the columns to insert.

    Returns:
        Range: Original cell.
    """

    match direction:
        case "down":
            return _move_down(sf, count)

        case "right":
            return _move_right(sf, count, width)

    raise ValueError("direction must be 'down' or 'right'")


def _move_down(sf: SheetFrame, count: int) -> None:
    start = sf.row - 1
    end = start + count - 1

    if sf.cell.offset(-1).formula:
        end += 1

    rows = sf.sheet.api.Rows(f"{start}:{end}")
    rows.Insert(Shift=Direction.xlDown)

    sf.cell = sf.cell.offset()  # update cell


def _move_right(sf: SheetFrame, count: int, width: int) -> None:
    start = sf.column - 1
    end = start + count - 1

    start_name = index_to_column_name(start)
    end_name = index_to_column_name(end)
    columns_name = f"{start_name}:{end_name}"

    columns = sf.sheet.api.Columns(columns_name)
    columns.Insert(Shift=Direction.xlToRight)

    if width:
        columns = sf.sheet.api.Columns(columns_name)
        columns.ColumnWidth = width

    sf.cell = sf.cell.offset()  # update cell



import pytest
from pandas import DataFrame, MultiIndex
from xlwings import Sheet

from xlviews.colors import rgb
from xlviews.config import rcParams
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.mark.parametrize(
    "name",
    ["index.name", "index", "columns.name", "columns", "values"],
)
def test_set_style(sheet: Sheet, name):
    from xlviews.dataframes.style import _set_style

    rng = sheet["C3:E5"]
    _set_style(rng[0], rng[-1], name)
    param = f"frame.{name}.fill.color"
    color = rgb(rcParams[param])
    assert rng.api.Interior.Color == color


@pytest.fixture(scope="module")
def sf_basic(sheet_module: Sheet):
    from xlviews.dataframes.style import set_frame_style

    df = DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    sf = SheetFrame(2, 2, data=df, sheet=sheet_module)
    set_frame_style(sf)
    return sf


@pytest.mark.parametrize(
    ("cell", "name"),
    [("B2", "index.name"), ("B6", "index"), ("D2", "columns"), ("D5", "values")],
)
def test_frame_style_basic(sf_basic: SheetFrame, cell: str, name: str):
    c = rgb(rcParams[f"frame.{name}.fill.color"])
    assert sf_basic.sheet[cell].api.Interior.Color == c


def test_frame_style_banding_succession(sheet_module: Sheet):
    from xlviews.dataframes.style import set_frame_style

    df = DataFrame({"x": [1, 1, 2, 2], "a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    df = df.set_index("x")
    sf = SheetFrame(2, 6, data=df, sheet=sheet_module)
    set_frame_style(sf, banding=True, succession=True)
    assert sf.sheet["F4"].api.FormatConditions(1)
    assert sf.sheet["H5"].api.FormatConditions(1)


@pytest.fixture(scope="module")
def df_mc():
    df_mc = DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
    mi = [("a", "b", "c"), ("d", "e", "f"), ("g", "h", "i")]
    df_mc.columns = MultiIndex.from_tuples(mi)
    df_mc.columns.names = ["x", "y", "z"]
    return df_mc


@pytest.fixture(scope="module")
def sf_mc(sheet_module: Sheet, df_mc: DataFrame):
    from xlviews.dataframes.style import set_frame_style

    sf = SheetFrame(9, 2, data=df_mc, sheet=sheet_module)
    set_frame_style(sf)
    return sf


@pytest.mark.parametrize(
    ("cell", "name"),
    [
        ("B9", "columns.name"),
        ("B11", "columns.name"),
        ("B12", "index"),
        ("B15", "index"),
        ("C9", "columns"),
        ("E10", "columns"),
        ("C11", "columns"),
        ("E11", "columns"),
        ("C12", "values"),
        ("E15", "values"),
    ],
)
def test_frame_style_multi_columns(sf_mc: SheetFrame, cell: str, name: str):
    c = rgb(rcParams[f"frame.{name}.fill.color"])
    assert sf_mc.sheet[cell].api.Interior.Color == c


@pytest.fixture(scope="module")
def df_mic(df_mc: DataFrame):
    df_mic = df_mc.copy()
    df_mic.columns = MultiIndex.from_tuples([("a", "b"), ("c", "d"), ("e", "f")])
    df_mic.columns.names = ["x", "y"]
    i = [("i", "j"), ("k", "l"), ("m", "n"), ("o", "p")]
    df_mic.index = MultiIndex.from_tuples(i)
    return df_mic


@pytest.fixture(scope="module")
def sf_mic(sheet_module: Sheet, df_mic: DataFrame):
    from xlviews.dataframes.style import set_frame_style

    sf = SheetFrame(9, 7, data=df_mic, sheet=sheet_module)
    set_frame_style(sf)
    return sf


@pytest.mark.parametrize(
    ("cell", "name"),
    [
        ("G9", "index.name"),
        ("H10", "index.name"),
        ("G11", "index"),
        ("H14", "index"),
        ("I9", "columns"),
        ("K9", "columns"),
        ("I10", "columns"),
        ("K10", "columns"),
        ("I11", "values"),
        ("K14", "values"),
    ],
)
def test_frame_style_multi_index_names(sf_mic: SheetFrame, cell: str, name: str):
    c = rgb(rcParams[f"frame.{name}.fill.color"])
    assert sf_mic.sheet[cell].api.Interior.Color == c


@pytest.fixture(scope="module")
def sf_wide(sheet_module: Sheet):
    from xlviews.dataframes.style import set_wide_column_style

    df = DataFrame({"x": ["i", "j"], "y": ["k", "l"], "a": [1, 2], "b": [3, 4]})
    sf = SheetFrame(24, 2, data=df, sheet=sheet_module)
    sf.add_wide_column("u", range(3), autofit=True)
    sf.add_wide_column("v", range(4), autofit=True)
    set_wide_column_style(sf)
    return sf


@pytest.mark.parametrize(
    ("cell", "name"),
    [("G23", ".name"), ("M23", ".name"), ("G24", ""), ("M24", "")],
)
def test_frame_style_wide(sf_wide: SheetFrame, cell: str, name: str):
    c = rgb(rcParams[f"frame.wide-columns{name}.fill.color"])
    assert sf_wide.sheet[cell].api.Interior.Color == c


def test_table_style(sheet_module: Sheet):
    from xlviews.dataframes.style import set_table_style

    df = DataFrame({"x": [1, 1, 2, 2], "a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    df = df.set_index("x")
    sf = SheetFrame(17, 2, data=df, sheet=sheet_module)
    table = sf.as_table(style=False)
    assert table
    set_table_style(table)
    assert table.sheet.book.api.TableStyles("xlviews")



import pytest
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import FrameContainer, is_app_available
from xlviews.testing.sheet_frame.base import WideColumn

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return WideColumn(sheet_module)


@pytest.fixture(scope="module")
def df(fc: FrameContainer):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: FrameContainer):
    return fc.sf


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.row == 3
    assert sf.column == 29
    assert sf.sheet.name == sheet_module.name
    assert sf.index.nlevels == 2
    assert sf.columns.nlevels == 1


def test_len(sf: SheetFrame):
    assert len(sf) == 5


def test_index_names(sf: SheetFrame):
    assert sf.index.names == ["x", "y"]


def test_contains(sf: SheetFrame):
    assert "x" not in sf
    assert "a" in sf
    assert "u" in sf


def test_iter(sf: SheetFrame):
    assert list(sf) == ["a", "b", "u", "v"]


def test_get_number_format(sf: SheetFrame):
    assert sf.number_format(u="0.00")
    assert sf.get_number_format("u") == "0.00"



from xlviews.chart.axes import Axes
from xlviews.chart.series import Series
from xlviews.core.formula import aggregate
from xlviews.core.range import Range
from xlviews.dataframes.colorbar import Colorbar
from xlviews.dataframes.dist_frame import DistFrame
from xlviews.dataframes.groupby import GroupBy
from xlviews.dataframes.heat_frame import HeatFrame
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.dataframes.stats_frame import StatsFrame
from xlviews.utils import constant

__all__ = [
    "Axes",
    "Colorbar",
    "DistFrame",
    "GroupBy",
    "HeatFrame",
    "Range",
    "Series",
    "SheetFrame",
    "StatsFrame",
    "aggregate",
    "constant",
]



import pytest
import xlwings
from xlwings import App, Book


@pytest.fixture(scope="session")
def app():
    with xlwings.App(visible=False) as app:
        yield app


@pytest.fixture(scope="session")
def book(app: App):
    book = app.books.add()

    yield book

    book.close()


@pytest.fixture(scope="module")
def sheet_module(book: Book):
    from xlviews.style import hide_gridlines

    sheet = book.sheets.add()
    hide_gridlines(sheet)

    yield sheet

    sheet.delete()


@pytest.fixture
def sheet(book: Book):
    sheet = book.sheets.add()

    yield sheet

    sheet.delete()



from __future__ import annotations

from typing import TYPE_CHECKING

from xlwings import Range as RangeImpl
from xlwings.constants import ChartType

from xlviews.colors import Color, rgb
from xlviews.core.range import Range
from xlviews.core.range_collection import RangeCollection

from .style import get_line_style, get_marker_style

if TYPE_CHECKING:
    from typing import Any, Self

    from .axes import Axes


class Series:
    axes: Axes
    api: Any

    def __init__(
        self,
        axes: Axes,
        x: Any,
        y: Any | None = None,
        label: str | None = None,
        chart_type: int | None = None,
    ) -> None:
        self.axes = axes
        self.api = axes.chart.api[1].SeriesCollection().NewSeries()
        self.label = label

        if chart_type is not None:
            self.chart_type = chart_type

        if y is not None:
            self.x = x
            self.y = y

        else:
            self.y = x

    @property
    def label(self) -> str:
        return self.api.Name

    @label.setter
    def label(self, label: str | None) -> None:
        self.api.Name = label or ""

    @property
    def chart_type(self) -> int:
        return self.api.ChartType  # type: ignore

    @chart_type.setter
    def chart_type(self, chart_type: int) -> None:
        self.api.ChartType = chart_type

    @property
    def x(self) -> tuple:
        return self.api.XValues  # type: ignore

    @x.setter
    def x(self, x: Any) -> None:
        if isinstance(x, Range | RangeImpl | RangeCollection):
            self.api.XValues = x.api
        else:
            self.api.XValues = x

    @property
    def y(self) -> tuple:
        return self.api.Values  # type: ignore

    @y.setter
    def y(self, y: Any) -> None:
        if isinstance(y, Range | RangeImpl | RangeCollection):
            self.api.Values = y.api
        else:
            self.api.Values = y

    def delete(self) -> None:
        self.api.Delete()

    def set(
        self,
        style: str | int | None = None,
        color: Color | None = None,
        alpha: float | None = None,
        weight: float | None = None,
        marker: str | int | None = None,
        size: int | None = None,
        line: str | None = None,
        label: str | None = None,
    ) -> Self:
        if label is not None:
            self.label = label

        if self.chart_type == ChartType.xlXYScatter:  # marker
            weight = weight or 1
            return self.marker(style or marker, size, color, alpha, weight)

        weight = weight or 2
        return self.line(style or line or "-", weight, color, alpha, marker, size)

    def marker(
        self,
        style: str | int | None = None,
        size: int | None = None,
        color: Color | None = None,
        alpha: float | None = None,
        weight: float | None = None,
    ) -> Self:
        set_marker(self.api, get_marker_style(style), size)

        if color is not None:
            set_fill(self.api, rgb(color), alpha)
            alpha = alpha / 2 if weight and alpha is not None else alpha
            set_line(self.api, get_line_style(""), rgb(color), alpha, weight)

        return self

    def line(
        self,
        style: str | int | None = None,
        weight: float | None = None,
        color: Color | None = None,
        alpha: float | None = None,
        marker: str | int | None = None,
        size: int | None = None,
    ) -> Self:
        if color is not None:
            color = rgb(color)

        set_line(self.api, get_line_style(style), color, alpha, weight)

        if marker:
            set_marker(self.api, get_marker_style(marker), size)
            if color is not None:
                set_fill(self.api, color, alpha)

        return self


def set_marker(api: Any, style: int | None, size: int | None) -> None:
    if style is not None:
        api.MarkerStyle = style
    if size is not None:
        api.MarkerSize = size


def set_fill(api: Any, color: int | None, alpha: float | None) -> None:
    if color is not None:
        api.Format.Fill.BackColor.RGB = color
        api.Format.Fill.ForeColor.RGB = color
    if alpha is not None:
        api.Format.Fill.Transparency = alpha


def set_line(
    api: Any,
    style: int | None,
    color: int | None,
    alpha: float | None,
    weight: float | None,
) -> None:
    if style is None:
        style = api.Border.LineStyle

    if weight is not None:
        api.Format.Line.Weight = weight
    if color is not None:
        api.Format.Line.ForeColor.RGB = color
    if alpha is not None:
        api.Format.Line.Transparency = alpha

    api.Border.LineStyle = style  # must be set after weight and color



import pytest
from xlwings import Sheet
from xlwings.constants import ChartType, TickMark

from xlviews.chart.axes import Axes
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture
def ax(sheet_module: Sheet):
    ct = ChartType.xlXYScatterLines
    return Axes(
        left=300,
        top=10,
        width=250,
        height=300,
        chart_type=ct,
        sheet=sheet_module,
    )


def test_chart_position(ax: Axes):
    assert ax.chart.left == 300
    assert ax.chart.top == 10


def test_chart_dimensions(ax: Axes):
    assert ax.chart.width == 250
    assert ax.chart.height == 300


def test_chart_type(ax: Axes):
    assert ax.chart.api[1].ChartType == ChartType.xlXYScatterLines
    assert ax.chart.chart_type == "xy_scatter_lines"


def test_chart_row_column(sheet_module: Sheet):
    ax = Axes(row=2, column=3, sheet=sheet_module)
    assert ax.chart.left == 104
    assert ax.chart.top == 18


@pytest.mark.parametrize("axis", ["xaxis", "yaxis"])
def test_axis(ax: Axes, axis: str):
    mark = TickMark.xlTickMarkInside
    assert getattr(ax, axis).MajorTickMark == mark



from __future__ import annotations

from itertools import product

import pandas as pd
from pandas import DataFrame

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing.common import FrameContainer, create_sheet


class Base(FrameContainer):
    @classmethod
    def dataframe(cls) -> DataFrame:
        values = list(product(range(1, 5), range(1, 7)))
        df = DataFrame(values, columns=["x", "y"])
        df["u"] = list(range(len(df), 2 * len(df)))
        df["v"] = list(range(len(df)))
        df = df[(df["x"] + df["y"]) % 4 != 0]
        return df.set_index(["x", "y"])


class MultiIndex(FrameContainer):
    column: int = 17

    @classmethod
    def dataframe(cls) -> DataFrame:
        df = Base.dataframe().reset_index()

        dfs = []
        for x in range(1, 4):
            for y in range(1, 5):
                a = df.copy()
                a["X"] = x
                a["Y"] = y
                dfs.append(a)

        df = pd.concat(dfs)
        df["u"] = list(range(len(df), 2 * len(df)))
        df["v"] = list(range(len(df)))
        return df.set_index(["X", "Y", "x", "y"])


class Pivot(FrameContainer):
    @classmethod
    def dataframe(cls) -> DataFrame:
        df = MultiIndex.dataframe().reset_index()

        for a, b in [(1, 3), (2, 2), (2, 4)]:
            df = df[~((df["X"] == a) & (df["x"] == b))]

        for a, b in [(2, 1), (3, 3), (4, 1), (4, 3)]:
            df = df[~((df["Y"] == a) & (df["y"] == b))]

        dfs = []
        for x in range(1, 3):
            for y in range(1, 4):
                if x == 1 and y == 3:
                    continue

                a = df.copy()
                a["A"] = x
                a["B"] = y
                dfs.append(a)

        df = pd.concat(dfs)

        for a, b in [(2, 1)]:
            df = df[~((df["A"] == a) & (df["X"] == b))]

        for a, b in [(2, 3)]:
            df = df[~((df["B"] == a) & (df["Y"] == b))]

        df["u"] = (
            1000 * df["A"] + 300 * df["B"] + 100 * df["X"] + 10 * df["Y"] + df["x"]
        )
        df["v"] = list(range(len(df)))
        return df.set_index(["A", "B", "X", "Y", "x", "y"])


if __name__ == "__main__":
    sheet = create_sheet()

    fc = Base(sheet, style=True)
    sf = fc.sf
    sf.set_adjacent_column_width(1)
    df = sf.pivot_table("u", "y", "x", formula=True)
    SheetFrame(2, 7, df).style()
    df = sf.pivot_table(None, "y", "x", formula=True)
    SheetFrame(10, 7, df).style().autofit()
    df = sf.pivot_table("u", None, "x", formula=True, aggfunc="mean")
    SheetFrame(19, 7, df).style().autofit()
    df = sf.pivot_table("v", "y", None, formula=True, aggfunc="max")
    SheetFrame(22, 7, df).style().autofit()

    sf.set_adjacent_column_width(1)
    fc = MultiIndex(sheet, style=True)
    sf = fc.sf
    sf.set_adjacent_column_width(1)

    df = sf.pivot_table("u", ["Y", "y"], ["X", "x"], formula=True)
    SheetFrame(2, 24, df).style()
    df = sf.pivot_table("v", "Y", ["X", "y", "x"], formula=True)
    SheetFrame(29, 24, df).style().autofit()
    df = sf.pivot_table(["u", "v"], ["Y", "y"], ["X", "x"], formula=True)
    SheetFrame(37, 24, df).style()



from __future__ import annotations

from typing import ClassVar

from pandas import DataFrame

from xlviews.dataframes.dist_frame import DistFrame
from xlviews.testing.common import FrameContainer, create_sheet


class Parent(FrameContainer):
    row: int = 3
    column: int = 2
    index: ClassVar[list[str] | str] = ["x", "y"]

    @classmethod
    def dataframe(cls) -> DataFrame:
        df = DataFrame(
            {
                "x": [1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2],
                "y": [3, 3, 3, 3, 3, 4, 4, 4, 4, 3, 3, 3, 4, 4],
                "a": [5, 4, 3, 2, 1, 4, 3, 2, 1, 3, 2, 1, 2, 1],
                "b": [10, 4, 9, 14, 5, 4, 6, 3, 4, 9, 12, None, 9, 2],
            },
        )
        return df.set_index(cls.index)


if __name__ == "__main__":
    sheet = create_sheet()
    Parent.index = ["x", "y"]
    fc = Parent(sheet, 3, 2, style=True)
    fc.sf.set_adjacent_column_width(1)
    fc.sf.number_format(b="0.0")
    DistFrame(fc.sf, ["a", "b"], by=["x", "y"])



import pytest
from pandas import DataFrame, Series
from xlwings import App, Sheet

from xlviews.core.range import Range
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import FrameContainer, is_app_available
from xlviews.testing.sheet_frame.base import NoIndex

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return NoIndex(sheet_module)


@pytest.fixture(scope="module")
def df(fc: FrameContainer):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: FrameContainer):
    return fc.sf


def test_init(sf: SheetFrame, fc: FrameContainer):
    assert sf.row == fc.row
    assert sf.column == fc.column
    assert sf.index.nlevels == 1
    assert sf.columns.nlevels == 1


def test_repr(sf: SheetFrame):
    assert repr(sf).endswith("!$B$2:$D$6>")


def test_str(sf: SheetFrame):
    assert str(sf).endswith("!$B$2:$D$6>")


def test_len(sf: SheetFrame):
    assert len(sf) == 4


@pytest.mark.parametrize(("x", "b"), [("a", True), ("x", False), (0, False)])
def test_contains(sf: SheetFrame, x, b):
    assert (x in sf) is b


def test_iter(sf: SheetFrame):
    assert list(sf) == ["a", "b"]


def test_value(sf: SheetFrame, df: DataFrame):
    df_sf = sf.value
    assert df_sf.equals(df.astype(float))
    assert df_sf.index.equals(df.index)
    assert df_sf.columns.equals(df.columns)


@pytest.mark.parametrize(("column", "loc"), [("a", 3), ("b", 4)])
def test_loc(sf: SheetFrame, column, loc):
    assert sf.get_loc(column) == loc


@pytest.mark.parametrize(
    ("columns", "indexer"),
    [(["a"], [3]), (["b"], [4]), (None, [3, 4])],
)
def test_get_indexer(sf: SheetFrame, columns, indexer):
    assert sf.get_indexer(columns) == indexer


@pytest.mark.parametrize(
    ("column", "offset", "address"),
    [
        ("a", 0, "$C$3"),
        ("a", -1, "$C$2"),
        ("b", -1, "$D$2"),
        ("a", None, "$C$3:$C$6"),
    ],
)
def test_get_range(sf: SheetFrame, column: str, offset, address):
    assert sf.get_range(column, offset).get_address() == address


def test_get_range_error(sf: SheetFrame):
    with pytest.raises(ValueError, match="invalid offset"):
        sf.get_range(None, offset=1)  # type: ignore


@pytest.mark.parametrize(
    ("axis", "v0", "v1"),
    [(0, [1, 2, 3, 4], [5, 6, 7, 8]), (1, [1, 5], [2, 6])],
)
def test_iter_ranges(sf: SheetFrame, axis, v0, v1):
    values = list(sf.iter_ranges(axis))
    assert values[0].value == v0
    assert values[1].value == v1


def test_iter_ranges_error(sf: SheetFrame):
    with pytest.raises(ValueError, match="axis must be 0 or 1"):
        list(sf.iter_ranges(axis=2))  # type: ignore


@pytest.mark.parametrize("columns", [["a", "b"], None])
def test_get_address_none(sf: SheetFrame, columns):
    df = sf.get_address(columns)
    assert df.columns.to_list() == ["a", "b"]
    assert df.index.to_list() == [0, 1, 2, 3]
    assert df.loc[0, "a"] == "$C$3"
    assert df.loc[0, "b"] == "$D$3"
    assert df.loc[1, "a"] == "$C$4"
    assert df.loc[1, "b"] == "$D$4"
    assert df.loc[2, "a"] == "$C$5"
    assert df.loc[2, "b"] == "$D$5"
    assert df.loc[3, "a"] == "$C$6"
    assert df.loc[3, "a"] == "$C$6"


def test_get_address_str(sf: SheetFrame):
    s = sf.get_address("a", formula=True, row_absolute=False)
    assert s.name == "a"
    assert s.index.to_list() == [0, 1, 2, 3]
    assert s.to_list() == ["=$C3", "=$C4", "=$C5", "=$C6"]


def test_agg_str(sf: SheetFrame):
    s = sf.agg("sum", row_absolute=False, column_absolute=False)
    assert s.name == "sum"
    assert s.index.to_list() == ["a", "b"]
    assert s.to_list() == ["AGGREGATE(9,7,C3:C6)", "AGGREGATE(9,7,D3:D6)"]


def test_agg_dict(sf: SheetFrame):
    s = sf.agg({"a": "min", "b": "max"}, row_absolute=False, column_absolute=False)
    assert s.name is None
    assert s.index.to_list() == ["a", "b"]
    assert s.to_list() == ["AGGREGATE(5,7,C3:C6)", "AGGREGATE(4,7,D3:D6)"]


def test_agg_none(sf: SheetFrame):
    s = sf.agg()
    assert s.name is None
    assert s.index.to_list() == ["a", "b"]
    assert s.to_list() == ["$C$3:$C$6", "$D$3:$D$6"]


def test_agg_range(sf: SheetFrame):
    func = Range(100, 200, sf.sheet)
    s = sf.agg(func)
    assert s.name is None
    assert s.index.to_list() == ["a", "b"]
    assert 'IF(GR100="soa"' in s["a"]


def test_agg_range_error_sheet(sf: SheetFrame, sheet: Sheet):
    func = Range(100, 200, sheet)
    with pytest.raises(ValueError, match="Range is from a different sheet"):
        sf.agg(func)


def test_agg_range_error_book(sf: SheetFrame, app: App):
    book = app.books.add()

    func = Range(100, 200, book.sheets[0])
    with pytest.raises(ValueError, match="Range is from a different book"):
        sf.agg(func)

    book.close()


def test_agg_list(sf: SheetFrame):
    df = sf.agg(["median", "mean"])
    assert df.index.to_list() == ["median", "mean"]
    assert df.columns.to_list() == ["a", "b"]
    assert df.loc["median", "a"] == "AGGREGATE(12,7,$C$3:$C$6)"
    assert df.loc["mean", "b"] == "AGGREGATE(1,7,$D$3:$D$6)"


def test_agg_first(sf: SheetFrame):
    s = sf.agg("first")
    assert s.name == "first"
    assert s.to_list() == ["$C$3", "$D$3"]


def test_agg_columns_str(sf: SheetFrame):
    s = sf.agg(None, "a")
    assert s.name is None
    assert s.index.to_list() == ["a"]
    assert s.to_list() == ["$C$3:$C$6"]


def test_melt_none(sf: SheetFrame):
    s = sf.melt()
    assert isinstance(s, Series)
    assert s.name is None
    assert s.index.to_list() == ["a", "b"]
    assert s.to_list() == ["$C$3:$C$6", "$D$3:$D$6"]


def test_melt_str(sf: SheetFrame):
    s = sf.melt("std")
    assert isinstance(s, Series)
    assert s.name == "std"
    assert s.index.to_list() == ["a", "b"]
    assert s.to_list() == ["AGGREGATE(8,7,$C$3:$C$6)", "AGGREGATE(8,7,$D$3:$D$6)"]



from __future__ import annotations

from typing import TYPE_CHECKING

from xlwings import Range
from xlwings.constants import AutoFilterOperator, ListObjectSourceType, YesNoGuess

from xlviews.core.formula import const
from xlviews.style import set_alignment, set_font

from .style import set_table_style

if TYPE_CHECKING:
    from xlwings import Sheet
    from xlwings._xlwindows import COMRetryObjectWrapper


class Table:
    sheet: Sheet
    cell: Range
    api: COMRetryObjectWrapper

    def __init__(
        self,
        rng: Range | None = None,
        *,
        autofit: bool = False,
        const_header: bool = False,
        style: bool = False,
        sheet: Sheet | None = None,
        api: COMRetryObjectWrapper | None = None,
        index_nlevels: int | None = None,
    ) -> None:
        if isinstance(rng, Range):
            self.cell = rng[0]

            self.api = rng.sheet.api.ListObjects.Add(
                ListObjectSourceType.xlSrcRange,
                rng.api,
                None,
                YesNoGuess.xlYes,
            )
        elif sheet and api:
            self.api = api
            self.cell = sheet.range(api.Range.Row, api.Range.Column)  # type: ignore
        else:
            raise ValueError("Either range or sheet and api must be provided")

        self.sheet = self.cell.sheet

        if autofit:
            self.header.api.EntireColumn.AutoFit()

        if const_header:
            self.add_const_header(index_nlevels)

        if style:
            set_table_style(self)

    @property
    def column(self) -> Range:
        start = self.cell.offset(1)
        end = start.expand("down")
        return self.cell.sheet.range(start, end)

    @property
    def header(self) -> Range:
        end = self.cell.expand("right")
        return self.cell.sheet.range(self.cell, end)

    @property
    def const_header(self) -> Range:
        return self.header.offset(-1)

    @property
    def columns(self) -> list[str]:
        names = self.header.value

        if not isinstance(names, list):
            raise NotImplementedError

        return [c or "" for c in names]

    def add_const_header(
        self,
        columns: int | None = None,
        *,
        clear: bool = False,
    ) -> None:
        """Write the filtered element above the header.

        Args:
            columns (int | None, optional): The number of columns to add the const
                header to.
            clear (bool, optional): If True, clear the header.
        """
        if clear:
            self.const_header.value = None
        else:
            const_header = self.const_header[:columns]
            if const_header:
                const_header.value = "=" + const(self.column)
                set_font(const_header, size=8, italic=True, color="blue")
                set_alignment(const_header, "center")

    def auto_filter(self, *args, clear: bool = False, **field_criteria) -> None:
        """Filter the table by the given criteria.

        Args:
            *args: The criteria to filter by.
            **field_criteria: The criteria to filter by.

        The criteria is a dictionary with the column name as the key
        and the criteria as the value. The criteria can be one of the following:
           - list: specify the elements.
           - tuple: specify the range of values.
           - None: clear the filter.
           - other: specify the value.
        """
        if clear:
            clear_filter = {x: None for x in self.columns}
            clear_filter.update(field_criteria)
            field_criteria = clear_filter

        if args and isinstance(args[0], dict):
            field_criteria.update(args[0])
        else:
            for name, criteria in zip(args[::2], args[1::2], strict=True):
                field_criteria[name] = criteria

        auto_filter = self.api.Range.AutoFilter  # type: ignore

        columns = self.columns

        for name, criteria in field_criteria.items():
            field = columns.index(name) + 1

            if isinstance(criteria, list):
                auto_filter(
                    Field=field,
                    Criteria1=[str(x) for x in criteria],
                    Operator=AutoFilterOperator.xlFilterValues,
                )

            elif isinstance(criteria, tuple):
                auto_filter(
                    Field=field,
                    Criteria1=f">={criteria[0]}",
                    Operator=AutoFilterOperator.xlAnd,
                    Criteria2=f"<={criteria[1]}",
                )

            elif criteria is None:
                auto_filter(Field=field)

            else:
                auto_filter(Field=field, Criteria1=f"{criteria}")

    def unlist(self) -> None:
        """Unlist the SheetFrame."""
        self.auto_filter(clear=True)
        self.api.Unlist()
        self.add_const_header(clear=True)



import pytest
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.figure.plot import Plot
from xlviews.testing import is_app_available
from xlviews.testing.figure.facet import facet
from xlviews.testing.sheet_frame.pivot import Pivot

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return Pivot(sheet_module)


@pytest.fixture(scope="module")
def df(fc: Pivot):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: Pivot):
    return fc.sf


@pytest.fixture(scope="module")
def plots(sf: SheetFrame):
    return [plot for _, plot in facet(sf)]


def test_plots_len(plots: list[Plot]):
    assert len(plots) == 5


@pytest.mark.parametrize(
    ("i", "title"),
    [(0, "1_1"), (1, "2_1"), (2, "1_2"), (3, "2_2"), (4, "2_3")],
)
def test_title(plots: list[Plot], i: int, title: str):
    assert plots[i].axes.title == title


def test_iterrows_none():
    from pandas import Index

    from xlviews.figure.plot import iterrows

    assert list(iterrows(Index([]), None)) == [{}]



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.groupby import groupby
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import FrameContainer, is_app_available
from xlviews.testing.sheet_frame.base import MultiIndex

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return MultiIndex(sheet_module)


@pytest.fixture(scope="module")
def df(fc: FrameContainer):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: FrameContainer):
    return fc.sf


def test_init(sf: SheetFrame, sheet_module: Sheet):
    assert sf.row == 2
    assert sf.column == 6
    assert sf.sheet.name == sheet_module.name
    assert sf.index.nlevels == 2
    assert sf.columns.nlevels == 1


def test_len(sf: SheetFrame):
    assert len(sf) == 8


def test_index_names(sf: SheetFrame):
    assert sf.index.names == ["x", "y"]


@pytest.mark.parametrize(("x", "b"), [("x", False), ("b", True)])
def test_contains(sf: SheetFrame, x, b):
    assert (x in sf) is b


def test_iter(sf: SheetFrame):
    assert list(sf) == ["a", "b"]


def test_value(sf: SheetFrame, df: DataFrame):
    df_sf = sf.value
    assert df_sf.equals(df.astype(float))
    assert df_sf.index.equals(df.index)
    assert df_sf.columns.equals(df.columns)


@pytest.mark.parametrize(
    ("column", "offset", "address"),
    [
        ("x", -1, "$F$2"),
        ("y", 0, "$G$3"),
        ("a", -1, "$H$2"),
        ("b", 0, "$I$3"),
        ("y", None, "$G$3:$G$10"),
    ],
)
def test_get_range(sf: SheetFrame, column, offset, address):
    assert sf.get_range(column, offset).get_address() == address


@pytest.mark.parametrize(
    ("by", "v1", "v2"),
    [
        ("x", [(3, 6)], [(7, 10)]),
        ("y", [(3, 4), (7, 8)], [(5, 6), (9, 10)]),
    ],
)
def test_groupby(sf: SheetFrame, by, v1, v2):
    g = groupby(sf, by)
    assert len(g) == 2
    assert g[(1,)] == v1
    assert g[(2,)] == v2


def test_groupby_list(sf: SheetFrame):
    g = groupby(sf, ["x", "y"])
    assert len(g) == 4
    assert g[(1, 1)] == [(3, 4)]
    assert g[(1, 2)] == [(5, 6)]
    assert g[(2, 1)] == [(7, 8)]
    assert g[(2, 2)] == [(9, 10)]


def test_get_address(sf: SheetFrame):
    df = sf.get_address(row_absolute=False, column_absolute=False, formula=True)
    assert df.columns.to_list() == ["a", "b"]
    assert df.index.names == ["x", "y"]
    assert df.to_numpy()[0, 0] == "=H3"



import pytest

from xlviews.config import rcParams


def test_config_file():
    from xlviews.config import CONFIG_FILE

    assert CONFIG_FILE.exists()


def test_rcparams():
    assert rcParams["chart.width"] == 200
    assert rcParams["chart.title.font.bold"] is True

    rcParams["chart.width"] = 100
    assert rcParams["chart.width"] == 100

    rcParams["chart.width"] = 200


def test_rcparams_error():
    with pytest.raises(KeyError):
        rcParams["invalid"]


def test_rcparams_get():
    assert rcParams.get("chart.width", 100) == 200


def test_rcparams_get_default():
    assert rcParams.get("invalid", "default") == "default"



def test_dist_func_str():
    from xlviews.dataframes.dist_frame import get_dist_func

    df = get_dist_func("norm", ["a", "b"])
    assert df == {"a": "norm", "b": "norm"}


def test_dist_func_dict():
    from xlviews.dataframes.dist_frame import get_dist_func

    df = get_dist_func({"a": "none"}, ["a", "b"])
    assert df == {"a": "none", "b": "norm"}



from .common import (
    FrameContainer,
    create_sheet,
    create_sheet_frame,
    is_app_available,
)

__all__ = [
    "FrameContainer",
    "create_sheet",
    "create_sheet_frame",
    "is_app_available",
]



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


def test_add_column(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    sf = SheetFrame(2, 2, data=df, sheet=sheet)
    sf.add_column("c", autofit=True, style=True)
    assert sf.columns.to_list() == ["a", "b", "c"]


def test_add_column_value(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    sf = SheetFrame(2, 2, data=df, sheet=sheet)
    sf.add_column("c", [7, 8, 9], number_format="0.0")
    assert sheet.range("E3:E5").value == [7, 8, 9]


@pytest.mark.parametrize(
    ("formula", "value"),
    [("={a}+{b}", [6, 8, 10, 12]), ("={a}*{b}", [5, 12, 21, 32])],
)
def test_add_formula_column(formula, value, sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    sf = SheetFrame(2, 3, data=df, sheet=sheet)
    sf.add_formula_column("c", formula)
    assert sheet.range("F3:F6").value == value

    sf.add_formula_column("c", formula + "+1", style=True)
    assert sheet.range("F3:F6").value == [v + 1 for v in value]


@pytest.mark.parametrize(
    ("formula", "value"),
    [
        ("={a}+{b}+{c}", ([7, 9, 11, 13], [10, 12, 14, 16])),
        ("={a}*{b}*{c}", ([5, 12, 21, 32], [20, 48, 84, 128])),
    ],
)
def test_formula_wide(formula, value, sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    sf = SheetFrame(10, 3, data=df, sheet=sheet)
    sf.add_wide_column("c", [1, 2, 3, 4], number_format="0", style=True)
    sf.add_formula_column("c", formula, number_format="0", autofit=True)
    assert sheet.range((11, 6), (14, 6)).value == value[0]
    assert sheet.range((11, 9), (14, 9)).value == value[1]



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.dist_frame import DistFrame
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import FrameContainer
from xlviews.testing.dist_frame import Parent


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return Parent(sheet_module, 3, 2, style=True)


@pytest.fixture(scope="module")
def df_parent(fc: FrameContainer):
    return fc.df


@pytest.fixture(scope="module")
def sf_parent(fc: FrameContainer):
    return fc.sf


@pytest.fixture(scope="module")
def sf(sf_parent: SheetFrame):
    return DistFrame(sf_parent, ["a", "b"], by=["x", "y"])


@pytest.fixture(scope="module")
def df(sf: DistFrame):
    rng = sf.expand().impl
    return rng.options(DataFrame, index=sf.index.nlevels).value



import pytest
from xlwings import Sheet
from xlwings.constants import ChartType

from xlviews.chart.axes import Axes
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def ax(sheet_module: Sheet):
    ct = ChartType.xlXYScatterLines
    ax = Axes(2, 2, ct, sheet_module)
    x = sheet_module["B2:B11"]
    y = sheet_module["C2:C11"]
    z = sheet_module["D2:D11"]
    x.options(transpose=True).value = list(range(1, 11))
    y.options(transpose=True).value = list(range(10, 20))
    z.options(transpose=True).value = list(range(20, 30))
    ax.add_series(x, y, label="a")
    ax.add_series(x, z, label="b")
    return ax


def test_title_str(ax: Axes):
    ax.title = "a"
    assert ax.title == "a"


def test_title_none(ax: Axes):
    ax.title = None
    assert ax.title is None


def test_title_sytle(ax: Axes):
    ax.set_title("Style", name="Times", size=20, bold=True, italic=True)
    assert ax.title == "Style"
    chart_title = ax.chart.api[1].ChartTitle
    assert chart_title.Font.Name == "Times"
    assert chart_title.Font.Size == 20
    assert chart_title.Font.Bold
    assert chart_title.Font.Italic


def test_xlabel_str(ax: Axes):
    ax.xlabel = "a"
    assert ax.xlabel == "a"


def test_xlabel_none(ax: Axes):
    ax.xlabel = None
    assert ax.xlabel is None


def test_ylabel_str(ax: Axes):
    ax.ylabel = "a"
    assert ax.ylabel == "a"


def test_ylabel_none(ax: Axes):
    ax.ylabel = None
    assert ax.ylabel is None


def test_xticks(ax: Axes):
    ax.xticks = (5, 40, 10, 2)
    assert ax.xticks == (5, 40, 10, 2)


def test_yticks(ax: Axes):
    ax.yticks = (10, 40, 10, 2)
    assert ax.yticks == (10, 40, 10, 2)


def test_xticks_set(ax: Axes):
    ax.set_xticks(2, 5, 2, 1, gridlines=False)
    assert ax.xticks == (2, 5, 2, 1)


def test_yticks_set(ax: Axes):
    ax.set_yticks(2, 7, 2, 1, gridlines=False)
    assert ax.yticks == (2, 7, 2, 1)


def test_xtick_labels(ax: Axes):
    ax.set_xtick_labels("Arial", size=10, number_format="0.00")
    assert ax.xaxis.TickLabels.Font.Name == "Arial"
    assert ax.xaxis.TickLabels.Font.Size == 10
    assert ax.xaxis.TickLabels.NumberFormatLocal == "0.00"


def test_ytick_labels(ax: Axes):
    ax.set_ytick_labels("Times", size=14, number_format="0.0")
    assert ax.yaxis.TickLabels.Font.Name == "Times"
    assert ax.yaxis.TickLabels.Font.Size == 14
    assert ax.yaxis.TickLabels.NumberFormatLocal == "0.0"


@pytest.mark.parametrize("scale", ["log", "linear"])
def test_xscale(ax: Axes, scale: str):
    ax.xscale = scale
    assert ax.xscale == scale
    ax.xscale = "linear"


@pytest.mark.parametrize("scale", ["log", "linear"])
def test_yscale(ax: Axes, scale: str):
    ax.yscale = scale
    assert ax.yscale == scale
    ax.yscale = "linear"


def test_set(ax: Axes):
    ax.set(
        xlabel="1",
        ylabel="2",
        xticks=(1, 10, 4, 1),
        yticks=(2, 20, 4, 1),
        xscale="linear",
        yscale="linear",
        title="Title",
        legend=True,
    )
    assert ax.xlabel == "1"
    assert ax.ylabel == "2"
    assert ax.xticks == (1, 10, 4, 1)
    assert ax.yticks == (2, 20, 4, 1)
    assert ax.xscale == "linear"
    assert ax.yscale == "linear"
    assert ax.title == "Title"


@pytest.fixture
def ax2(sheet: Sheet):
    ct = ChartType.xlXYScatterLines
    ax = Axes(2, 2, ct, sheet)
    x = sheet["B2:B11"]
    y = sheet["C2:C11"]
    z = sheet["D2:D11"]
    x.options(transpose=True).value = list(range(1, 11))
    y.options(transpose=True).value = list(range(10, 20))
    z.options(transpose=True).value = list(range(20, 30))
    ax.add_series(x, y, label="a")
    ax.add_series(x, z, label="b")
    return ax


@pytest.mark.parametrize(
    ("loc", "left", "top"),
    [
        ((-1, -1), 25, 148),
        ((-1, 1), 25, 10),
        ((0, 0), 85, 80),
        ((1, 1), 145, 10),
        ((1, -1), 145, 148),
    ],
)
def test_legend_position(ax2: Axes, loc, left, top):
    ax2.set(legend=loc)
    assert ax2.chart.api[1].HasLegend
    legend = ax2.chart.api[1].Legend
    assert left - 2 < legend.Left < left + 2
    assert top - 2 < legend.Top < top + 2


def test_tight_layout(ax: Axes):
    ax.xlabel = "x"
    ax.ylabel = "y"
    ax.title = "title"
    ax.tight_layout()
    assert 0 <= ax.chart.api[1].ChartTitle.Top <= 10

    ax.xlabel = None
    ax.tight_layout()


def test_style(ax: Axes):
    ax.style()
    assert ax.chart.api[1].PlotArea.Format.Line.Visible


def test_legend_none(sheet: Sheet):
    ct = ChartType.xlXYScatterLines
    ax = Axes(2, 2, ct, sheet)
    x = sheet["B2:B11"]
    y = sheet["C2:C11"]
    z = sheet["D2:D11"]
    x.options(transpose=True).value = list(range(1, 11))
    y.options(transpose=True).value = list(range(10, 20))
    z.options(transpose=True).value = list(range(20, 30))
    ax.add_series(x, y)
    ax.add_series(x, z)
    ax.legend()
    assert ax.chart.api[1].HasLegend is False



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture
def sf(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    return SheetFrame(3, 3, df, sheet)


def test_number_format_kwargs(sf: SheetFrame):
    sf.number_format(autofit=False, a="0", b="0.0")
    assert sf.sheet.range("D4:D7").number_format == "0"
    assert sf.sheet.range("E4:E7").number_format == "0.0"


def test_number_format_dict(sf: SheetFrame):
    sf.number_format({r"[ab]": "0.00"}, autofit=True)
    assert sf.get_number_format("a") == "0.00"
    assert sf.get_number_format("b") == "0.00"


@pytest.mark.parametrize(
    ("alignment", "value"),
    [("left", -4131), ("center", -4108), ("right", -4152)],
)
def test_alignment(sf: SheetFrame, alignment: str, value: int):
    sf.alignment(alignment)
    assert sf.cell.api.HorizontalAlignment == value


def test_adjacent_column_width(sf: SheetFrame):
    sf.set_adjacent_column_width(10)
    assert sf.sheet["F1"].column_width == 10


# def test_child_frame(sf: SheetFrame):
#     cell = sf.get_child_cell()
#     assert cell.get_address() == "$G$3"

#     cell = sf.get_adjacent_cell(offset=3)
#     assert cell.get_address() == "$J$3"

#     df = DataFrame({"x": [1, 2], "y": [5, 6], "z": [7, 8]})
#     sf_child = SheetFrame(parent=sf, data=df)
#     assert sf_child.cell.get_address() == "$G$3"

#     assert sf_child.parent is sf
#     assert sf.children[0] is sf_child

#     cell = sf.get_adjacent_cell()
#     assert cell.get_address() == "$L$3"


# def test_head_frame(sf: SheetFrame):
#     df = DataFrame({"x": [1, 2], "y": [5, 6], "z": [7, 8]})
#     sf_tail = SheetFrame(head=sf, data=df)
#     assert sf_tail.cell.get_address() == "$C$9"
#     assert sf.tail is sf_tail
#     assert sf_tail.head is sf


def test_active_sheet(sheet: Sheet):
    sheet.name = "active"
    sheet.activate()
    df = DataFrame({"x": [1, 2], "y": [5, 6], "z": [7, 8]})
    sf = SheetFrame(100, 10, data=df)
    assert sf.sheet.name == "active"
    assert sf.cell.get_address() == "$J$100"



from __future__ import annotations

from typing import TYPE_CHECKING

import xlwings

from xlviews.colors import rgb
from xlviews.config import rcParams
from xlviews.core.formula import aggregate
from xlviews.core.range import Range
from xlviews.style import (
    set_alignment,
    set_border,
    set_color_scale,
    set_font,
    set_number_format,
)

if TYPE_CHECKING:
    from typing import Literal, Self

    from xlwings import Sheet


class Colorbar:
    start: int
    end: int
    offset: int
    orientation: Literal["vertical", "horizontal"] = "vertical"
    sheet: Sheet
    range: Range

    def __init__(
        self,
        row: int,
        column: int,
        length: int,
        orientation: Literal["vertical", "horizontal"] = "vertical",
        sheet: Sheet | None = None,
    ) -> None:
        self.sheet = sheet or xlwings.sheets.active
        self.orientation = orientation

        if orientation == "vertical":
            self.start = row
            self.end = row + length - 1
            self.offset = column
            self.range = Range((self.start, column), (self.end, column), self.sheet)

        else:
            self.start = column
            self.end = column + length - 1
            self.offset = row
            self.range = Range((row, self.start), (row, self.end), self.sheet)

    def set(
        self,
        vmin: float | str | Range | list[Range] | None = None,
        vmax: float | str | Range | list[Range] | None = None,
        label: str | None = None,
        autofit: bool = False,
    ) -> Self:
        if vmin is not None:
            self.vmin = vmin
        if vmax is not None:
            self.vmax = vmax
        if label is not None:
            self.label = label

        self.draw()

        if autofit:
            self.autofit()

        return self

    @property
    def vmin(self) -> Range:
        i = -1 if self.orientation == "vertical" else 0
        return self.range[i]

    @property
    def vmax(self) -> Range:
        i = 0 if self.orientation == "vertical" else -1
        return self.range[i]

    @vmin.setter
    def vmin(self, value: float | str | Range | list[Range]) -> None:
        if isinstance(value, Range | list):
            func = "min" if len(value) > 1 else None
            value = aggregate(func, value, formula=True)

        self.vmin.value = value

    @vmax.setter
    def vmax(self, value: float | str | Range | list[Range]) -> None:
        if isinstance(value, Range | list):
            func = "max" if len(value) > 1 else None
            value = aggregate(func, value, formula=True)

        self.vmax.value = value

    @property
    def label(self) -> Range:
        offset = (-1, 0) if self.orientation == "vertical" else (0, 1)
        return self.vmax.offset(*offset)

    @label.setter
    def label(self, label: str | None) -> None:
        rng = self.label
        rng.value = label
        set_font(rng, bold=True, size=rcParams["frame.font.size"])
        set_alignment(rng, horizontal_alignment="center")

    def draw(self) -> None:
        rng = self.range
        set_color_scale(rng, self.vmin, self.vmax)
        set_font(rng, color=rgb("white"), size=rcParams["frame.font.size"])
        set_alignment(rng, horizontal_alignment="center")
        ec = rcParams["heat.border.color"]
        set_border(rng, edge_weight=2, edge_color=ec, inside_weight=0)

        vmin = self.vmin.get_address()
        vmax = self.vmax.get_address()

        n = self.end - self.start - 1
        for i in range(n):
            value = f"={vmax}+{i + 1}*({vmin}-{vmax})/{n + 1}"
            if self.orientation == "vertical":
                rng = self.sheet.range(self.start + i + 1, self.offset)
            else:
                rng = self.sheet.range(self.offset, self.start + i + 1)

            rng.value = value
            set_font(rng, size=4)
            set_number_format(rng, "0")

    def apply(self, rng: Range) -> None:
        set_color_scale(rng, self.vmin, self.vmax)

    def autofit(self) -> Self:
        if self.orientation == "vertical":
            start = (self.start - 1, self.offset)
            end = (self.end, self.offset)
        else:
            start = (self.offset, self.start)
            end = (self.offset, self.end + 1)

        self.sheet.range(start, end).autofit()
        return self

    def set_adjacent_column_width(self, width: float, offset: int = 1) -> None:
        """Set the width of the adjacent empty column."""
        if self.orientation == "vertical":
            self.range.offset(0, 1).impl.column_width = width
        else:
            self.range.last_cell.offset(0, 2).impl.column_width = width



import numpy as np
import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.core.address import index_to_column_name
from xlviews.dataframes.groupby import GroupBy
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


def create_data_frame(rows: int = 10, columns: int = 20) -> DataFrame:
    values = np.arange(rows * columns).reshape((rows, columns))
    cnames = [index_to_column_name(i + 1) for i in range(columns)]
    df = DataFrame(values, columns=cnames)
    return df.set_index(df.columns.to_list()[:10])


def create_sheet_frame(df: DataFrame, sheet: Sheet) -> SheetFrame:
    return SheetFrame(2, 3, data=df, sheet=sheet)


@pytest.mark.parametrize(("rows", "columns"), [(10, 20), (100, 100), (1000, 100)])
def test_create_sheet_frame(benchmark, sheet: Sheet, rows: int, columns: int):
    df = create_data_frame(rows, columns)
    sf = benchmark(create_sheet_frame, df, sheet)
    assert isinstance(sf, SheetFrame)


@pytest.fixture(
    params=[(100, 20), (1000, 20), (10000, 20), (10, 100), (10, 1000)],
    ids=lambda x: "_".join([str(i) for i in x]),
)
def shape(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture
def sf(shape: tuple[int, int], sheet: Sheet):
    rows, columns = shape
    df = create_data_frame(rows, columns)
    return create_sheet_frame(df, sheet)


@pytest.fixture(
    params=[
        ["A"],
        ["A", "B", "C", "D"],
        ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J"],
    ],
    ids=lambda x: f"C{len(x)}",
)
def columns(request: pytest.FixtureRequest):
    return request.param


@pytest.mark.parametrize("axis", [0, 1])
def test_ranges(benchmark, sf: SheetFrame, shape, axis):
    x = benchmark(lambda: list(sf.iter_ranges(axis)))
    assert len(x) == shape[1 - axis] - 10 * (1 - axis)


def test_agg(benchmark, sf: SheetFrame, columns):
    x = benchmark(lambda: sf.agg(["sum", "count"], columns))
    assert isinstance(x, DataFrame)
    assert x.shape == (2, len(columns))


def test_groupby(benchmark, sf: SheetFrame, columns):
    x = benchmark(lambda: sf.groupby(columns))
    assert isinstance(x, GroupBy)
    assert len(x) == len(sf)


def test_groupby_agg(benchmark, sf: SheetFrame, columns, shape):
    x = benchmark(lambda: sf.groupby(columns).agg(["sum", "count"]))
    assert isinstance(x, DataFrame)
    assert x.shape == (len(sf), 2 * (shape[1] - 10))



from __future__ import annotations

from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import Literal

    from xlviews.chart.axes import Axes


class Series:
    axes: list[Axes]

    def __init__(
        self,
        ax: Axes | list[Axes],
        n: int = 0,
        *,
        axis: Literal[0, 1] = 0,
    ) -> None:
        if isinstance(ax, list):
            self.axes = ax
            return

        if axis == 0:
            series = Grid(ax, 1, n)[0, :]
        elif axis == 1:
            series = Grid(ax, n, 1)[:, 0]
        else:
            msg = f"Invalid axis: {axis}"
            raise ValueError(msg)

        self.axes = list(series)

    @overload
    def __getitem__(self, key: int) -> Axes: ...

    @overload
    def __getitem__(self, key: slice) -> Series: ...

    def __getitem__(self, key: int | slice) -> Axes | Series:
        if isinstance(key, int):
            return self.axes[key]

        if isinstance(key, slice):
            return Series(self.axes[key])

        msg = f"Invalid key: {key}"
        raise ValueError(msg)

    def __len__(self) -> int:
        return len(self.axes)

    def __iter__(self) -> Iterator[Axes]:
        return iter(self.axes)


class Grid:
    axes: list[list[Axes]]

    def __init__(
        self,
        ax: Axes | list[list[Axes]],
        nrows: int = 0,
        ncols: int = 0,
    ) -> None:
        if isinstance(ax, list):
            self.axes = ax
            return

        left = ax.chart.left
        top = ax.chart.top
        width = ax.chart.width
        height = ax.chart.height

        axes = []
        for r in range(nrows):
            row = []
            for c in range(ncols):
                if r == 0 and c == 0:
                    row.append(ax)
                else:
                    new = ax.copy(left=left + c * width, top=top + r * height)
                    row.append(new)
            axes.append(row)
        self.axes = axes

    @property
    def shape(self) -> tuple[int, int]:
        if self.axes:
            return len(self.axes), len(self.axes[0])
        return 0, 0

    @overload
    def __getitem__(self, key: int) -> Series: ...

    @overload
    def __getitem__(self, key: tuple[int, int]) -> Axes: ...

    @overload
    def __getitem__(self, key: tuple[int, int]) -> Axes: ...

    @overload
    def __getitem__(self, key: tuple[slice, int]) -> Series: ...

    @overload
    def __getitem__(self, key: tuple[int, slice]) -> Series: ...

    @overload
    def __getitem__(self, key: tuple[slice, slice]) -> Grid: ...

    def __getitem__(self, key: int | tuple) -> Axes | Series | Grid:
        if isinstance(key, int):
            return Series(self.axes[key])

        if isinstance(key, tuple) and len(key) == 2:
            r, c = key
            if isinstance(r, int) and isinstance(c, int):
                return self.axes[r][c]

            if isinstance(r, slice) and isinstance(c, int):
                return Series([row[c] for row in self.axes[r]])

            if isinstance(r, int) and isinstance(c, slice):
                return Series(self.axes[r][c])

            if isinstance(r, slice) and isinstance(c, slice):
                return Grid([row[c] for row in self.axes[r]])

        msg = f"Invalid key: {key}"
        raise ValueError(msg)

    def __len__(self) -> int:
        return len(self.axes)

    def __iter__(self) -> Iterator[Series]:
        for row in self.axes:
            yield Series(row)



import pytest
from xlwings import Sheet
from xlwings.constants import ChartType

from xlviews.chart.axes import Axes
from xlviews.chart.series import Series
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def ax(sheet_module: Sheet):
    ct = ChartType.xlXYScatterLines
    return Axes(300, 10, chart_type=ct, sheet=sheet_module)


@pytest.fixture(scope="module")
def series(ax: Axes):
    return ax.add_series(
        [1, 2, 3],
        [4, 5, 6],
        label="xy",
        chart_type=ChartType.xlXYScatterLines,
    )


@pytest.fixture
def series_y(ax: Axes):
    series = ax.add_series([7, 8, 9], label="y")
    yield series
    series.delete()


def test_series_x(series: Series):
    assert series.x == ("1", "2", "3")


def test_series_y(series: Series):
    assert series.y == (4, 5, 6)


def test_series_y_y(series_y: Series):
    assert series_y.y == (7, 8, 9)


def test_chart_type(series: Series):
    assert series.chart_type == ChartType.xlXYScatterLines


def test_name(series: Series):
    assert series.label == "xy"



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.testing import is_app_available


@pytest.mark.parametrize(
    ("name", "value"),
    [
        ("ChartType.xlXYScatter", -4169),
        ("BordersIndex.EdgeTop", 8),
        ("Bottom", -4107),
        ("Center", -4108),
        ("Left", -4131),
        ("None", -4142),
        ("Right", -4152),
        ("Top", -4160),
    ],
)
def test_constant(name: str, value: int):
    from xlviews.utils import constant

    assert constant(name) == value
    assert constant(*name.split(".")) == value


@pytest.mark.parametrize(
    ("columns", "lst"),
    [
        ("B", ["B"]),
        (["A", "C"], ["A", "C"]),
        (":B", ["A", "B"]),
        (["::B", "C"], ["A", "C"]),
    ],
)
def test_iter_columns(columns, lst):
    from xlviews.utils import iter_columns

    df = DataFrame([[1, 2, 3], [4, 5, 6], [7, 8, 9]], columns=["A", "B", "C"])
    assert list(iter_columns(df, columns)) == lst


@pytest.mark.skipif(not is_app_available(), reason="Excel not installed")
def test_validate_list(sheet: Sheet):
    from xlviews.utils import add_validate_list

    rng = sheet.range("a1")
    add_validate_list(rng, [1, 2, 3], 2)
    assert rng.value == 2

    assert rng.api.Validation.Type == 3
    assert rng.api.Validation.Operator == 3
    assert rng.api.Validation.Formula1 == "1,2,3"



import pytest
from xlwings import Sheet

from xlviews.core.range_collection import RangeCollection
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.mark.parametrize(
    ("row", "n", "address"),
    [
        ([(4, 5), (10, 14)], 2, "E4:E5,E10:E14"),
        ([(5, 5), (7, 8), (10, 11)], 3, "E5,E7:E8,E10:E11"),
        ([1, 5], 2, "E1,E5"),
        (4, 1, "E4"),
    ],
)
def test_range_collection_row(row, n, address, sheet_module: Sheet):
    rc = RangeCollection(row, 5, sheet_module)
    assert len(rc) == n
    a = rc.get_address(row_absolute=False, column_absolute=False)
    assert a == address


@pytest.mark.parametrize(
    ("column", "n", "address"),
    [
        ([(2, 2)], 1, "$B$5"),
        ([(4, 5), (10, 14)], 2, "$D$5:$E$5,$J$5:$N$5"),
        ([(5, 5), (7, 8), (10, 11)], 3, "$E$5,$G$5:$H$5,$J$5:$K$5"),
        ([1, 5], 2, "$A$5,$E$5"),
        (4, 1, "$D$5"),
    ],
)
def test_range_collection_column(column, n, address, sheet_module: Sheet):
    rc = RangeCollection(5, column, sheet_module)
    assert len(rc) == n
    assert rc.get_address() == address
    assert rc.api.Address == address


def test_range_collection_iter(sheet_module: Sheet):
    rc = RangeCollection([(2, 5), (10, 12)], 1, sheet_module)
    for rng, row in zip(rc, [2, 10], strict=True):
        assert rng.row == row


def test_range_collection_repr(sheet_module: Sheet):
    rc = RangeCollection([(2, 5), (8, 10)], 5, sheet_module)
    assert repr(rc) == "<RangeCollection $E$2:$E$5,$E$8:$E$10>"


def test_iter_ranges_error(sheet_module: Sheet):
    with pytest.raises(TypeError):
        RangeCollection((1, 2), (3, 4), sheet_module)



import numpy as np
import pytest
from pandas import DataFrame

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.dataframes.stats_frame import StatsFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def sf(sf_parent: SheetFrame):
    funcs = ["count", "max", "median", "soa"]
    sf = StatsFrame(sf_parent, funcs, by=":y")
    sf.as_table()
    return sf


@pytest.fixture(scope="module")
def df(sf: StatsFrame):
    rng = sf.expand().impl
    return rng.options(DataFrame, index=sf.index.nlevels).value


def test_len(sf: StatsFrame):
    assert len(sf) == 16


def test_index_names(sf: StatsFrame):
    assert sf.index.names == ["func", "x", "y", "z"]


@pytest.mark.parametrize(
    ("func", "column", "value"),
    [
        ("count", "a", [7, 3, 4, 4]),
        ("count", "b", [8, 4, 4, 4]),
        ("count", "c", [7, 3, 3, 4]),
        ("max", "a", [18, 7, 11, 15]),
        ("max", "b", [27, 7, 9, 15]),
        ("max", "c", [24, 34, 36, 10]),
        ("median", "a", [3, 6, 9.5, 13.5]),
        ("median", "b", [10.5, 5.5, 5.5, 10.5]),
        ("median", "c", [18, 30, 2, 7]),
    ],
)
def test_value_float(df: DataFrame, func, column, value):
    np.testing.assert_allclose(df.loc[func][column], value)


@pytest.mark.parametrize(
    ("column", "value"),
    [
        ("a", [[0, 1, 2, 3, 16, 17, 18], [5, 6, 7]]),
        ("b", [[0, 1, 2, 3, 18, 21, 24, 27], [4, 5, 6, 7]]),
        ("c", [[20, 22, 24, 12, 14, 16, 18], [28, 30, 34]]),
    ],
)
def test_value_soa(df: DataFrame, column, value):
    soa = [np.std(x) / np.median(x) for x in value]
    np.testing.assert_allclose(df.loc["soa"][column][:2], soa)



import numpy as np
import pytest
from scipy.stats import norm
from xlwings import Sheet

from xlviews.core.range import Range
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(params=["norm", "weibull"])
def dist(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture
def value(sheet_module: Sheet, dist: str):
    from xlviews.dataframes.dist_frame import set_formula, sigma_value

    rng = Range((1, 1), (10, 1), sheet_module)
    rng.value = [[x] for x in range(1, 11)]
    cell = rng[0]
    sigma = sigma_value(cell, 10, dist)
    set_formula(cell.offset(0, 1), 10, sigma)
    return Range((1, 2), (10, 2), sheet_module).value


@pytest.fixture
def expected(dist: str):
    if dist == "norm":
        return norm.ppf(np.arange(1, 11) / 11)

    return np.log(-np.log(1 - np.arange(1, 11) / 11))


def test_sigma_value(value, expected):
    np.testing.assert_allclose(value, expected)


def test_sigma_value_error(sheet_module: Sheet):
    from xlviews.dataframes.dist_frame import sigma_value

    with pytest.raises(ValueError, match="unknown distribution"):
        sigma_value(Range((1, 1), (1, 1), sheet_module), 10, "unknown")



import pytest
from xlwings import Sheet

from xlviews.dataframes.heat_frame import HeatFrame
from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available
from xlviews.testing.heat_frame.facet import facet
from xlviews.testing.sheet_frame.pivot import Pivot

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc(sheet_module: Sheet):
    return Pivot(sheet_module)


@pytest.fixture(scope="module")
def df(fc: Pivot):
    return fc.df


@pytest.fixture(scope="module")
def sf(fc: Pivot):
    return fc.sf


@pytest.fixture(scope="module")
def hfs(sf: SheetFrame):
    return [hf for _, hf in facet(sf)]


@pytest.mark.parametrize(
    ("i", "v"),
    [(0, 1411), (1, 2511), (2, 1711), (3, 2811), (4, None), (5, 3111)],
)
def test_value(hfs: list[HeatFrame], i: int, v):
    assert hfs[i].cell.offset(1, 1).value == v



"""Set styles for Range."""

from __future__ import annotations

from typing import TYPE_CHECKING

import xlwings
from xlwings import Range as RangeImpl
from xlwings import Sheet
from xlwings.constants import (
    BordersIndex,
    ConditionValueTypes,
    FormatConditionType,
    LineStyle,
)

from xlviews.colors import Color, rgb
from xlviews.config import rcParams
from xlviews.core.range import Range
from xlviews.utils import constant

if TYPE_CHECKING:
    from xlwings._xlwindows import COMRetryObjectWrapper

    from .core.range_collection import RangeCollection


def set_border_line(
    rng: Range | RangeImpl,
    index: str,
    weight: int = 2,
    color: Color = 0,
) -> None:
    if not weight:
        return

    borders = rng.api.Borders
    border = borders(getattr(BordersIndex, index))
    border.LineStyle = LineStyle.xlContinuous
    border.Weight = weight
    border.Color = rgb(color)


def set_border_edge(
    rng: Range | RangeImpl,
    weight: int | tuple[int, int, int, int] = 3,
    color: Color = 0,
) -> None:
    if isinstance(weight, int):
        wl = wr = wt = wb = weight
    else:
        wl, wr, wt, wb = weight

    sheet = rng.sheet
    start, end = rng[0], rng[-1]

    left = sheet.range((start.row, start.column - 1), (end.row, start.column))
    set_border_line(left, "xlInsideVertical", weight=wl, color=color)

    right = sheet.range((start.row, end.column), (end.row, end.column + 1))
    set_border_line(right, "xlInsideVertical", weight=wr, color=color)

    top = sheet.range((start.row - 1, start.column), (start.row, end.column))
    set_border_line(top, "xlInsideHorizontal", weight=wt, color=color)

    bottom = sheet.range((end.row, start.column), (end.row + 1, end.column))
    set_border_line(bottom, "xlInsideHorizontal", weight=wb, color=color)


def set_border_inside(
    rng: Range | RangeImpl,
    weight: int = 1,
    color: Color = 0,
) -> None:
    set_border_line(rng, "xlInsideVertical", weight=weight, color=color)
    set_border_line(rng, "xlInsideHorizontal", weight=weight, color=color)


def set_border(
    rng: Range | RangeImpl,
    edge_weight: int | tuple[int, int, int, int] = 2,
    inside_weight: int = 1,
    edge_color: Color = 0,
    inside_color: Color = 0,
) -> None:
    if edge_weight:
        set_border_edge(rng, edge_weight, edge_color)

    if inside_weight:
        set_border_inside(rng, inside_weight, inside_color)


def set_fill(
    rng: Range | RangeCollection | RangeImpl,
    color: Color | None = None,
) -> None:
    if color is not None:
        rng.api.Interior.Color = rgb(color)


def set_font_api(
    api: COMRetryObjectWrapper,
    name: str | None = None,
    *,
    size: float | None = None,
    bold: bool | None = None,
    italic: bool | None = None,
    color: Color | None = None,
) -> None:
    font = api.Font
    if name:
        font.Name = name  # type: ignore
    if size:
        font.Size = size  # type: ignore
    if bold is not None:
        font.Bold = bold  # type: ignore
    if italic is not None:
        font.Italic = italic  # type: ignore
    if color is not None:
        font.Color = rgb(color)  # type: ignore


def set_font(
    rng: Range | RangeCollection | RangeImpl,
    name: str | None = None,
    *,
    size: float | None = None,
    bold: bool | None = None,
    italic: bool | None = None,
    color: Color | None = None,
) -> None:
    name = name or rcParams["frame.font.name"]
    set_font_api(rng.api, name, size=size, bold=bold, italic=italic, color=color)


def set_alignment(
    rng: Range | RangeCollection | RangeImpl,
    horizontal_alignment: str | None = None,
    vertical_alignment: str | None = None,
) -> None:
    if horizontal_alignment:
        rng.api.HorizontalAlignment = constant(horizontal_alignment)

    if vertical_alignment:
        rng.api.VerticalAlignment = constant(vertical_alignment)


def set_number_format(rng: Range | RangeCollection | RangeImpl, fmt: str) -> None:
    rng.api.NumberFormat = fmt


EVEN_COLOR = rgb(240, 250, 255)
ODD_COLOR = rgb(255, 255, 255)


def set_banding(
    rng: Range | RangeImpl,
    axis: int = 0,
    even_color: Color = EVEN_COLOR,
    odd_color: Color = ODD_COLOR,
) -> None:
    def banding(mod: int, color: int) -> None:
        formula = f"=MOD(ROW(), 2)={mod}" if axis == 0 else f"=MOD(COLUMN(), 2)={mod}"
        condition = add(Type=FormatConditionType.xlExpression, Formula1=formula)

        condition.SetFirstPriority()
        condition.StopIfTrue = False

        interior = condition.Interior
        interior.PatternColorIndex = constant("automatic")
        interior.Color = color
        interior.TintAndShade = 0

    add = rng.api.FormatConditions.Add

    banding(0, rgb(odd_color))
    banding(1, rgb(even_color))


SUCCESSION_COLOR = rgb(200, 200, 200)


def hide_succession(
    rng: Range | RangeImpl,
    color: Color = SUCCESSION_COLOR,
) -> None:
    cell = rng[0].get_address(row_absolute=False, column_absolute=False)

    start = rng[0].offset(-2).get_address(column_absolute=False)
    column = rng[0].offset(-1)
    column = ":".join(
        [
            column.get_address(column_absolute=False),
            column.get_address(row_absolute=False, column_absolute=False),
        ],
    )

    ref = (
        f"INDIRECT(ADDRESS(MAX(INDEX(SUBTOTAL(3,OFFSET({start},"
        f'ROW(INDIRECT("1:"&ROWS({column}))),))*ROW({column}),)),'
        f"COLUMN({column})))"
    )
    formula = f"={cell}={ref}"

    add = rng.api.FormatConditions.Add
    condition = add(Type=FormatConditionType.xlExpression, Formula1=formula)
    condition.SetFirstPriority()
    condition.StopIfTrue = False
    condition.Font.Color = rgb(color)


UNIQUE_COLOR = rgb(100, 100, 100)


def hide_unique(
    rng: Range | RangeImpl,
    length: int,
    color: Color = UNIQUE_COLOR,
) -> None:
    def address(r: Range | RangeImpl) -> str:
        return r.get_address(row_absolute=False, column_absolute=False)

    start = (rng[0].row + 1, rng[0].column)
    end = (rng[0].row + length, rng[0].column)
    cell = address(Range(start, end))
    ref = address(Range(start))
    formula = f"=COUNTIF({cell}, {ref}) = {length}"

    add = rng.api.FormatConditions.Add
    condition = add(Type=FormatConditionType.xlExpression, Formula1=formula)
    condition.SetFirstPriority()
    condition.StopIfTrue = False
    condition.Font.Color = rgb(color)
    condition.Font.Italic = True


def hide_gridlines(sheet: Sheet | None = None) -> None:
    sheet = sheet or xlwings.sheets.active
    sheet.book.app.api.ActiveWindow.DisplayGridlines = False


def set_color_condition(
    rng: Range | RangeImpl,
    values: list[str],
    colors: list[int],
) -> None:
    condition = rng.api.FormatConditions.AddColorScale(len(values))
    condition.SetFirstPriority()

    for k, (value, color) in enumerate(zip(values, colors, strict=True)):
        criteria = condition.ColorScaleCriteria(k + 1)
        criteria.Type = ConditionValueTypes.xlConditionValueNumber
        criteria.Value = value
        criteria.FormatColor.Color = color


def set_color_scale(
    rng: Range | RangeImpl,
    vmin: float | str | Range | RangeImpl,
    vmax: float | str | Range | RangeImpl,
) -> None:
    if isinstance(vmin, Range | RangeImpl):
        vmin = vmin.get_address()

    if isinstance(vmax, Range | RangeImpl):
        vmax = vmax.get_address()

    values = [f"={vmin}", f"=({vmin} + {vmax}) / 2", f"={vmax}"]
    colors = [rgb(130, 130, 255), rgb(80, 185, 80), rgb(255, 130, 130)]

    set_color_condition(rng, values, colors)



import pytest
from xlwings import Sheet

from xlviews.chart.axes import Axes
from xlviews.config import rcParams
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")

LEFT = rcParams["chart.left"]
TOP = rcParams["chart.top"]


def test_chart_position_none(sheet: Sheet):
    axes = Axes(sheet=sheet)
    assert axes.chart.left == LEFT
    assert axes.chart.top == TOP


def test_chart_position_from_cell(sheet: Sheet):
    axes = Axes(5, 10, sheet=sheet)
    assert axes.chart.left == 9 * sheet.cells(1, 1).width
    assert axes.chart.top == 4 * sheet.cells(1, 1).height


def test_chart_position_right(sheet: Sheet):
    a = Axes(sheet=sheet)
    b = Axes(sheet=sheet)
    assert b.chart.left == LEFT + a.chart.width
    assert a.chart.top == TOP


def test_chart_position_bottom(sheet: Sheet):
    a = Axes(sheet=sheet)
    b = Axes(sheet=sheet, top=-1)
    assert b.chart.left == LEFT
    assert b.chart.top == TOP + a.chart.height


def test_chart_position_left(sheet: Sheet):
    a = Axes(sheet=sheet)
    Axes(sheet=sheet)
    c = Axes(sheet=sheet, left=0)
    assert c.chart.left == LEFT
    assert c.chart.top == a.chart.top + a.chart.height


def test_chart_position_left_top(sheet: Sheet):
    Axes(sheet=sheet)
    Axes(sheet=sheet)
    c = Axes(sheet=sheet, left=0, top=100)
    assert c.chart.left == LEFT
    assert c.chart.top == 100


def test_chart_position_top(sheet: Sheet):
    a = Axes(sheet=sheet)
    Axes(sheet=sheet, top=-1)
    c = Axes(sheet=sheet, top=0)
    assert c.chart.left == a.chart.left + a.chart.width
    assert c.chart.top == TOP


def test_chart_position_top_left(sheet: Sheet):
    Axes(sheet=sheet)
    Axes(sheet=sheet, top=-1)
    c = Axes(sheet=sheet, top=0, left=200)
    assert c.chart.left == 200
    assert c.chart.top == TOP


def test_copy_right(sheet: Sheet):
    a = Axes(sheet=sheet)
    b = a.copy(left=0)
    assert b.chart.left == a.chart.left + a.chart.width


def test_copy_bottom(sheet: Sheet):
    a = Axes(sheet=sheet)
    b = a.copy(top=0)
    assert b.chart.top == a.chart.top + a.chart.height



import pytest
from pandas import DataFrame
from xlwings import Sheet

from xlviews.dataframes.sheet_frame import SheetFrame
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture
def sf(sheet: Sheet):
    df = DataFrame({"a": [1, 2, 3, 4], "b": [5, 6, 7, 8]})
    return SheetFrame(2, 3, data=df, sheet=sheet)


def test_range_original(sf: SheetFrame):
    assert sf.expand().get_address() == "$C$2:$E$6"


def test_move_down(sf: SheetFrame):
    sf.cell.offset(-1).value = "=$A$1"
    sf.move(20, "down")
    assert sf.cell.get_address() == "$C$23"
    assert sf.expand().get_address() == "$C$23:$E$27"
    assert sf.row == 23
    assert sf.column == 3
    v = [[None, "a", "b"], [0, 1, 5], [1, 2, 6], [2, 3, 7], [3, 4, 8]]
    assert sf.cell.expand().options(ndim=2).value == v
    assert sf.sheet.range("$C$2:$E$6").value == [[None] * 3] * 5


def test_move_right(sf: SheetFrame):
    sf.move(10, "right", width=3)
    assert sf.cell.get_address() == "$M$2"
    assert sf.expand().get_address() == "$M$2:$O$6"
    v = [[None, "a", "b"], [0, 1, 5], [1, 2, 6], [2, 3, 7], [3, 4, 8]]
    assert sf.cell.expand().options(ndim=2).value == v
    assert sf.sheet.range("$C$2:$E$6").value == [[None] * 3] * 5
    assert sf.sheet.range("C2").column_width == 3


def test_move_error(sf: SheetFrame):
    with pytest.raises(ValueError, match="direction must be 'down' or 'right'"):
        sf.move(10, "up")



import numpy as np
import pytest
from xlwings import Sheet

from xlviews.dataframes.heat_frame import HeatFrame
from xlviews.testing import is_app_available
from xlviews.testing.heat_frame.base import Base, BaseParent

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.fixture(scope="module")
def fc_parent(sheet_module: Sheet):
    return BaseParent(sheet_module)


@pytest.fixture(scope="module")
def sf(fc_parent: BaseParent):
    fc = Base(fc_parent.sf)
    return fc.sf


def test_index(sf: HeatFrame):
    assert sf.sheet.range("G3:G8").value == [1, 2, 3, 4, 5, 6]


def test_index_from_sf(sf: HeatFrame):
    assert sf.index.to_list() == [1, 2, 3, 4, 5, 6]


def test_columns(sf: HeatFrame):
    assert sf.sheet.range("H2:K2").value == [1, 2, 3, 4]


def test_columns_from_sf(sf: HeatFrame):
    assert sf.columns.to_list() == [1, 2, 3, 4]


@pytest.mark.parametrize(
    ("i", "value"),
    [
        (3, [0, 6, None, 18]),
        (4, [1, None, 13, 19]),
        (5, [None, 8, 14, 20]),
        (6, [3, 9, 15, None]),
        (7, [4, 10, None, 22]),
        (8, [5, None, 17, 23]),
    ],
)
def test_values(sf: HeatFrame, i: int, value: int):
    assert sf.sheet.range(f"H{i}:K{i}").value == value


def test_label(sf: HeatFrame):
    assert sf.sheet.range("M2").value == "v"


@pytest.mark.parametrize(
    ("i", "value"),
    [
        (3, 23),
        (4, 23 * 4 / 5),
        (5, 23 * 3 / 5),
        (6, 23 * 2 / 5),
        (7, 23 / 5),
        (8, 0),
    ],
)
def test_colorbar(sf: HeatFrame, i: int, value: int):
    v = sf.sheet.range(f"M{i}").value
    assert isinstance(v, float)
    np.testing.assert_allclose(v, value)


def test_number_format(sf: HeatFrame):
    sf.number_format("0.00")
    assert sf.sheet.range("H4").api.NumberFormatLocal == "0.00"
    assert sf.sheet.range("K8").api.NumberFormatLocal == "0.00"



import pytest
from xlwings import Sheet
from xlwings.constants import BordersIndex

from xlviews.config import rcParams
from xlviews.core.range import Range
from xlviews.testing import is_app_available

pytestmark = pytest.mark.skipif(not is_app_available(), reason="Excel not installed")


@pytest.mark.parametrize(("index", "value"), [("Vertical", 11), ("Horizontal", 12)])
def test_border_index(index, value):
    assert getattr(BordersIndex, f"xlInside{index}") == value


@pytest.fixture(params=[lambda x: x, Range.from_range], ids=["RangeImpl", "Range"])
def cls(request: pytest.FixtureRequest):
    return request.param


@pytest.mark.parametrize(
    ("weight", "value"),
    [((1, 2, 3, 4), (1, 2, -4138, 4)), (2, (2, 2, 2, 2))],
)
def test_border_edge(weight, value, cls, sheet: Sheet):
    from xlviews.style import set_border_edge

    set_border_edge(cls(sheet["C3:E5"]), weight, color="red")
    assert sheet["B3:C5"].api.Borders(11).Weight == value[0]
    assert sheet["B3:C5"].api.Borders(11).Color == 255
    assert sheet["E3:F5"].api.Borders(11).Weight == value[1]
    assert sheet["E3:F5"].api.Borders(11).Color == 255
    assert sheet["C2:E3"].api.Borders(12).Weight == value[2]
    assert sheet["C2:E3"].api.Borders(12).Color == 255
    assert sheet["C5:E6"].api.Borders(12).Weight == value[3]
    assert sheet["C5:E6"].api.Borders(12).Color == 255


def test_border_inside(cls, sheet: Sheet):
    from xlviews.style import set_border_inside

    set_border_inside(cls(sheet["C3:E5"]), weight=2, color="red")
    assert sheet["C3:E5"].api.Borders(11).Weight == 2
    assert sheet["C3:E5"].api.Borders(11).Color == 255
    assert sheet["C3:E5"].api.Borders(12).Weight == 2
    assert sheet["C3:E5"].api.Borders(12).Color == 255


def test_border(cls, sheet: Sheet):
    from xlviews.style import set_border

    set_border(cls(sheet["C3:E5"]), edge_weight=2, inside_weight=1)
    assert sheet["B3:C5"].api.Borders(11).Weight == 2
    assert sheet["C3:E5"].api.Borders(11).Weight == 1


def test_border_zero(cls, sheet: Sheet):
    from xlviews.style import set_border_line

    set_border_line(cls(sheet["C3:D5"]), "xlInsideVertical", weight=0, color="red")
    assert sheet["C3:D5"].api.Borders(11).Weight == 2
    assert sheet["C3:D5"].api.Borders(12).Weight == 2


def test_fill(cls, sheet: Sheet):
    from xlviews.style import set_fill

    set_fill(cls(sheet["C3:E5"]), color="pink")
    assert sheet["C3:E5"].api.Interior.Color == 13353215


def test_font(cls, sheet: Sheet):
    from xlviews.style import set_font

    rng = cls(sheet["C3"])
    rng.value = "abc"
    set_font(rng, "Times", size=24, bold=True, italic=True, color="green")
    assert rng.api.Font.Name == "Times"
    assert rng.api.Font.Size == 24
    assert rng.api.Font.Bold == 1
    assert rng.api.Font.Italic == 1
    assert rng.api.Font.Color == 32768


def test_font_collection(sheet: Sheet):
    from xlviews.core.range_collection import RangeCollection
    from xlviews.style import set_font

    rc = RangeCollection([(2, 3), (6, 7)], 2, sheet)
    set_font(rc, "Times", size=24, bold=True, italic=True, color="green")

    for row in [2, 3, 6, 7]:
        rng = sheet.range(row, 2)
        assert rng.api.Font.Name == "Times"
        assert rng.api.Font.Size == 24
        assert rng.api.Font.Bold == 1
        assert rng.api.Font.Italic == 1
        assert rng.api.Font.Color == 32768

    for row in [4, 5]:
        rng = sheet.range(row, 2)
        assert rng.api.Font.Size != 24


def test_font_without_name(cls, sheet: Sheet):
    from xlviews.style import set_font

    rng = cls(sheet["C3"])
    rng.value = "abc"
    set_font(rng)
    assert rng.api.Font.Name == rcParams["frame.font.name"]


@pytest.mark.parametrize(
    ("align", "value"),
    [("right", -4152), ("left", -4131), ("center", -4108)],
)
def test_alignment_horizontal(align, value, cls, sheet: Sheet):
    from xlviews.style import set_alignment

    rng = cls(sheet["C3"])
    rng.value = "a"
    set_alignment(rng, horizontal_alignment=align)
    assert rng.api.HorizontalAlignment == value


@pytest.mark.parametrize(
    ("align", "value"),
    [("top", -4160), ("bottom", -4107), ("center", -4108)],
)
def test_alignment_vertical(align, value, cls, sheet: Sheet):
    from xlviews.style import set_alignment

    rng = cls(sheet["C3"])
    rng.value = "a"
    set_alignment(rng, vertical_alignment=align)
    assert rng.api.VerticalAlignment == value


def test_number_format(cls, sheet: Sheet):
    from xlviews.style import set_number_format

    rng = cls(sheet["C3"])
    set_number_format(rng, "0.0%")
    assert rng.api.NumberFormat == "0.0%"


def test_number_format_collection(sheet: Sheet):
    from xlviews.core.range_collection import RangeCollection
    from xlviews.style import set_number_format

    rc = RangeCollection([(2, 3), (6, 7)], 3, sheet)
    set_number_format(rc, "0.0%")

    for row in [2, 3, 6, 7]:
        rng = sheet.range(row, 3)
        assert rng.api.NumberFormat == "0.0%"

    for row in [4, 5]:
        rng = sheet.range(row, 3)
        assert rng.api.NumberFormat != "0.0%"


@pytest.mark.parametrize(
    ("axis", "even_color", "odd_color"),
    [(0, 100, 200), (1, 300, 400)],
)
def test_banding(axis, even_color, odd_color, cls, sheet: Sheet):
    from xlviews.style import set_banding

    rng = cls(sheet["C3:F6"])
    set_banding(rng, axis, even_color, odd_color)
    assert rng.api.FormatConditions(1).Interior.Color == even_color
    assert rng.api.FormatConditions(2).Interior.Color == odd_color


def test_hide_succession(cls, sheet: Sheet):
    from xlviews.style import hide_succession

    rng = sheet["C3:C8"]
    rng.options(transpose=True).value = [1, 1, 2, 2, 3, 3]
    rng = sheet["D3:D8"]
    rng.options(transpose=True).value = [1, 1, 1, 2, 2, 2]
    rng = cls(sheet["C3:D8"])
    hide_succession(rng, color="red")
    assert rng.api.FormatConditions(1).Font.Color == 255


def test_hide_unique(cls, sheet: Sheet):
    from xlviews.style import hide_unique

    rng = sheet["C3:C8"]
    rng.options(transpose=True).value = [1, 1, 2, 2, 3, 3]
    rng = sheet["D3:D8"]
    rng.options(transpose=True).value = [1, 1, 1, 1, 1, 1]
    rng = cls(sheet["C2:D2"])
    rng.value = ["a", "b"]
    hide_unique(rng, 6, color="red")
    assert rng.api.FormatConditions(1).Font.Color == 255


def test_hide_gridlines(sheet: Sheet):
    from xlviews.style import hide_gridlines

    hide_gridlines(sheet)
    assert sheet.book.app.api.ActiveWindow.DisplayGridlines is False
